#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from designateclient import exceptions as designate_exception

from heat.common import exception
from heat.engine.resources.openstack.designate import zone
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils


sample_template = {
    'heat_template_version': '2015-04-30',
    'resources': {
        'test_resource': {
            'type': 'OS::Designate::Zone',
            'properties': {
                'name': 'test-zone.com',
                'description': 'Test zone',
                'ttl': 3600,
                'attributes': {
                    'tier': 'gold', 'ha': 'true'
                    },
                'email': 'abc@test-zone.com',
                'type': 'PRIMARY',
                'masters': []
            }
        }
    }
}


class DesignateZoneTest(common.HeatTestCase):

    def setUp(self):
        super(DesignateZoneTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(sample_template)
        )

        self.test_resource = self.stack['test_resource']

        # Mock client plugin
        self.test_client_plugin = mock.MagicMock()
        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

        # Mock client
        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

    def _get_mock_resource(self):
        value = {}
        value['id'] = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
        value['serial'] = '1434596972'

        return value

    def test_resource_handle_create(self):
        mock_zone_create = self.test_client.zones.create
        mock_resource = self._get_mock_resource()
        mock_zone_create.return_value = mock_resource

        # validate the properties
        self.assertEqual(
            'test-zone.com',
            self.test_resource.properties.get(zone.DesignateZone.NAME))
        self.assertEqual(
            'Test zone',
            self.test_resource.properties.get(
                zone.DesignateZone.DESCRIPTION))
        self.assertEqual(
            3600,
            self.test_resource.properties.get(zone.DesignateZone.TTL))
        self.assertEqual(
            'abc@test-zone.com',
            self.test_resource.properties.get(zone.DesignateZone.EMAIL))

        self.assertEqual(
            'PRIMARY',
            self.test_resource.properties.get(zone.DesignateZone.TYPE))

        self.assertEqual(
            [],
            self.test_resource.properties.get(zone.DesignateZone.MASTERS))

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()

        args = dict(
            name='test-zone.com',
            description='Test zone',
            ttl=3600,
            email='abc@test-zone.com',
            type_='PRIMARY',
            attributes=dict(tier='gold', ha='true')
        )

        mock_zone_create.assert_called_once_with(**args)
        # validate physical resource id
        self.assertEqual(mock_resource['id'], self.test_resource.resource_id)

    def _mock_check_status_active(self):
        self.test_client.zones.get.side_effect = [
            {'status': 'PENDING'},
            {'status': 'ACTIVE'},
            {'status': 'ERROR'}
        ]

    def test_check_create_complete(self):
        self._mock_check_status_active()
        self.assertFalse(self.test_resource.check_create_complete())
        self.assertTrue(self.test_resource.check_create_complete())
        ex = self.assertRaises(exception.ResourceInError,
                               self.test_resource.check_create_complete)
        self.assertIn('Error in zone',
                      ex.message)

    def _test_resource_validate(self, type_, prp):
        def _side_effect(key):
            if key == prp:
                return None
            if key == zone.DesignateZone.TYPE:
                return type_
            else:
                return sample_template['resources'][
                    'test_resource']['properties'][key]

        self.test_resource.properties = mock.MagicMock()
        self.test_resource.properties.get.side_effect = _side_effect
        self.test_resource.properties.__getitem__.side_effect = _side_effect

        ex = self.assertRaises(exception.StackValidationFailed,
                               self.test_resource.validate)
        self.assertEqual('Property %s is required for zone type %s' %
                         (prp, type_),
                         ex.message)

    def test_resource_validate_primary(self):
        self._test_resource_validate(zone.DesignateZone.PRIMARY,
                                     zone.DesignateZone.EMAIL)

    def test_resource_validate_secondary(self):
        self._test_resource_validate(zone.DesignateZone.SECONDARY,
                                     zone.DesignateZone.MASTERS)

    def test_resource_handle_update(self):
        mock_zone_update = self.test_client.zones.update
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {zone.DesignateZone.EMAIL: 'xyz@test-zone.com',
                     zone.DesignateZone.DESCRIPTION: 'updated description',
                     zone.DesignateZone.TTL: 4200}

        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        args = dict(
            description='updated description',
            ttl=4200,
            email='xyz@test-zone.com'
        )
        mock_zone_update.assert_called_once_with(
            self.test_resource.resource_id,
            args)

    def test_check_update_complete(self):
        self._mock_check_status_active()
        self.assertFalse(self.test_resource.check_update_complete())
        self.assertTrue(self.test_resource.check_update_complete())
        ex = self.assertRaises(exception.ResourceInError,
                               self.test_resource.check_update_complete)
        self.assertIn('Error in zone',
                      ex.message)

    def test_check_delete_complete(self):
        self.test_client.zones.get.side_effect = [
            {'status': 'PENDING'},
            designate_exception.NotFound,
            {'status': 'ERROR'}
        ]
        self.assertFalse(self.test_resource.check_delete_complete(
            self._get_mock_resource()['id']
        ))
        self.assertTrue(self.test_resource.check_delete_complete(
            self._get_mock_resource()['id']
        ))
        ex = self.assertRaises(exception.ResourceInError,
                               self.test_resource.check_delete_complete,
                               self._get_mock_resource()['id'])
        self.assertIn('Error in zone',
                      ex.message)

    def test_resolve_attributes(self):
        mock_zone = self._get_mock_resource()
        self.test_resource.resource_id = mock_zone['id']
        self.test_client.zones.get.return_value = mock_zone
        self.assertEqual(
            mock_zone['serial'],
            self.test_resource._resolve_attribute(zone.DesignateZone.SERIAL))
        self.test_client.zones.get.assert_called_once_with(
            self.test_resource.resource_id
        )


class DesignateSecondaryZoneTest(common.HeatTestCase):

    def setUp(self):
        super(DesignateSecondaryZoneTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.primaries = ['::1']

    def _get_mock_resource(self):
        value = {}
        value['id'] = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
        value['serial'] = '1434596972'

        return value

    def test_masters(self):
        self.do_test({
            'heat_template_version': '2015-04-30',
            'resources': {
                'test_resource': {
                    'type': 'OS::Designate::Zone',
                    'properties': {
                        'name': 'test-zone.com',
                        'description': 'Test zone',
                        'ttl': 3600,
                        'email': 'abc@test-zone.com',
                        'type': 'SECONDARY',
                        'masters': self.primaries,
                    }
                }
            }
        })

    def test_primaries(self):
        self.do_test({
            'heat_template_version': '2015-04-30',
            'resources': {
                'test_resource': {
                    'type': 'OS::Designate::Zone',
                    'properties': {
                        'name': 'test-zone.com',
                        'description': 'Test zone',
                        'ttl': 3600,
                        'email': 'abc@test-zone.com',
                        'type': 'SECONDARY',
                        'primaries': self.primaries,
                    }
                }
            }
        })

    def do_test(self, sample_template):
        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(sample_template)
        )

        self.test_resource = self.stack['test_resource']

        # Mock client plugin
        self.test_client_plugin = mock.MagicMock()
        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

        # Mock client
        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

        mock_zone_create = self.test_client.zones.create
        mock_resource = self._get_mock_resource()
        mock_zone_create.return_value = mock_resource

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()

        args = dict(
            name='test-zone.com',
            description='Test zone',
            ttl=3600,
            email='abc@test-zone.com',
            type_='SECONDARY',
            masters=self.primaries
        )

        mock_zone_create.assert_called_once_with(**args)
        # validate physical resource id
        self.assertEqual(mock_resource['id'], self.test_resource.resource_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import json
from unittest import mock

from oslo_utils import timeutils

from heat.common import exception
from heat.common import grouputils
from heat.common import template_format
from heat.engine.clients.os import nova
from heat.engine import resource
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests.autoscaling import inline_templates
from heat.tests import common
from heat.tests.openstack.nova import fakes as fakes_nova
from heat.tests import utils


as_template = inline_templates.as_template


class TestAutoScalingGroupValidation(common.HeatTestCase):
    def validate_scaling_group(self, t, stack, resource_name):
        # create the launch configuration resource
        conf = stack['LaunchConfig']
        self.assertIsNone(conf.validate())
        scheduler.TaskRunner(conf.create)()
        self.assertEqual((conf.CREATE, conf.COMPLETE), conf.state)

        # create the group resource
        rsrc = stack[resource_name]
        self.assertIsNone(rsrc.validate())
        return rsrc

    def test_toomany_vpc_zone_identifier(self):
        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['VPCZoneIdentifier'] = ['xxxx', 'yyyy']

        stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.stub_SnapshotConstraint_validate()
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()
        self.assertRaises(exception.NotSupported,
                          self.validate_scaling_group, t,
                          stack, 'WebServerGroup')

    def test_invalid_min_size(self):
        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['MinSize'] = '-1'
        properties['MaxSize'] = '2'

        stack = utils.parse_stack(t, params=inline_templates.as_params)

        self.stub_SnapshotConstraint_validate()
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()

        e = self.assertRaises(exception.StackValidationFailed,
                              self.validate_scaling_group, t,
                              stack, 'WebServerGroup')

        expected_msg = "The size of AutoScalingGroup can not be less than zero"
        self.assertEqual(expected_msg, str(e))

    def test_invalid_max_size(self):
        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['MinSize'] = '3'
        properties['MaxSize'] = '1'

        stack = utils.parse_stack(t, params=inline_templates.as_params)

        self.stub_SnapshotConstraint_validate()
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()

        e = self.assertRaises(exception.StackValidationFailed,
                              self.validate_scaling_group, t,
                              stack, 'WebServerGroup')

        expected_msg = "MinSize can not be greater than MaxSize"
        self.assertEqual(expected_msg, str(e))

    def test_invalid_desiredcapacity(self):
        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['MinSize'] = '1'
        properties['MaxSize'] = '3'
        properties['DesiredCapacity'] = '4'

        stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.stub_SnapshotConstraint_validate()
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()
        e = self.assertRaises(exception.StackValidationFailed,
                              self.validate_scaling_group, t,
                              stack, 'WebServerGroup')

        expected_msg = "DesiredCapacity must be between MinSize and MaxSize"
        self.assertEqual(expected_msg, str(e))

    def test_invalid_desiredcapacity_zero(self):
        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['MinSize'] = '1'
        properties['MaxSize'] = '3'
        properties['DesiredCapacity'] = '0'

        stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.stub_SnapshotConstraint_validate()
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()

        e = self.assertRaises(exception.StackValidationFailed,
                              self.validate_scaling_group, t,
                              stack, 'WebServerGroup')

        expected_msg = "DesiredCapacity must be between MinSize and MaxSize"
        self.assertEqual(expected_msg, str(e))

    def test_validate_without_InstanceId_and_LaunchConfigurationName(self):
        t = template_format.parse(as_template)
        agp = t['Resources']['WebServerGroup']['Properties']
        agp.pop('LaunchConfigurationName')
        agp.pop('LoadBalancerNames')
        stack = utils.parse_stack(t, params=inline_templates.as_params)
        rsrc = stack['WebServerGroup']
        error_msg = ("Either 'InstanceId' or 'LaunchConfigurationName' "
                     "must be provided.")
        exc = self.assertRaises(exception.StackValidationFailed,
                                rsrc.validate)
        self.assertIn(error_msg, str(exc))

    def test_validate_with_InstanceId_and_LaunchConfigurationName(self):
        t = template_format.parse(as_template)
        agp = t['Resources']['WebServerGroup']['Properties']
        agp['InstanceId'] = '5678'
        stack = utils.parse_stack(t, params=inline_templates.as_params)
        rsrc = stack['WebServerGroup']
        error_msg = ("Either 'InstanceId' or 'LaunchConfigurationName' "
                     "must be provided.")
        exc = self.assertRaises(exception.StackValidationFailed,
                                rsrc.validate)
        self.assertIn(error_msg, str(exc))

    def _stub_nova_server_get(self, not_found=False):
        mock_server = mock.MagicMock()
        mock_server.image = {'id': 'dd619705-468a-4f7d-8a06-b84794b3561a'}
        mock_server.flavor = {'id': '1'}
        mock_server.key_name = 'test'
        mock_server.security_groups = [{'name': 'hth_test'}]
        if not_found:
            self.patchobject(nova.NovaClientPlugin, 'get_server',
                             side_effect=exception.EntityNotFound(
                                 entity='Server', name='5678'))
        else:
            self.patchobject(nova.NovaClientPlugin, 'get_server',
                             return_value=mock_server)

    def test_scaling_group_create_with_instanceid(self):
        t = template_format.parse(as_template)
        agp = t['Resources']['WebServerGroup']['Properties']
        agp['InstanceId'] = '5678'
        agp.pop('LaunchConfigurationName')
        agp.pop('LoadBalancerNames')
        stack = utils.parse_stack(t, params=inline_templates.as_params)
        rsrc = stack['WebServerGroup']

        self._stub_nova_server_get()

        _config, ins_props = rsrc._get_conf_properties()

        self.assertEqual('dd619705-468a-4f7d-8a06-b84794b3561a',
                         ins_props['ImageId'])
        self.assertEqual('test', ins_props['KeyName'])
        self.assertEqual(['hth_test'], ins_props['SecurityGroups'])
        self.assertEqual('1', ins_props['InstanceType'])

    def test_scaling_group_create_with_instanceid_not_found(self):
        t = template_format.parse(as_template)
        agp = t['Resources']['WebServerGroup']['Properties']
        agp.pop('LaunchConfigurationName')
        agp['InstanceId'] = '5678'
        stack = utils.parse_stack(t, params=inline_templates.as_params)
        rsrc = stack['WebServerGroup']
        self._stub_nova_server_get(not_found=True)
        msg = ("Property error: "
               "Resources.WebServerGroup.Properties.InstanceId: "
               "Error validating value '5678': The Server (5678) could "
               "not be found.")
        exc = self.assertRaises(exception.StackValidationFailed,
                                rsrc.validate)
        self.assertIn(msg, str(exc))


class TestScalingGroupTags(common.HeatTestCase):
    def setUp(self):
        super(TestScalingGroupTags, self).setUp()
        t = template_format.parse(as_template)
        self.stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.group = self.stack['WebServerGroup']

    def test_tags_default(self):
        expected = [{'Key': 'metering.groupname',
                     'Value': 'WebServerGroup'},
                    {'Key': 'metering.AutoScalingGroupName',
                     'Value': 'WebServerGroup'}]
        self.assertEqual(expected, self.group._tags())

    def test_tags_with_extra(self):
        self.group.properties.data['Tags'] = [
            {'Key': 'fee', 'Value': 'foo'}]
        expected = [{'Key': 'fee',
                     'Value': 'foo'},
                    {'Key': 'metering.groupname',
                     'Value': 'WebServerGroup'},
                    {'Key': 'metering.AutoScalingGroupName',
                     'Value': 'WebServerGroup'}]
        self.assertEqual(expected, self.group._tags())

    def test_tags_with_metering(self):
        self.group.properties.data['Tags'] = [
            {'Key': 'metering.fee', 'Value': 'foo'}]
        expected = [{'Key': 'metering.fee', 'Value': 'foo'},
                    {'Key': 'metering.AutoScalingGroupName',
                     'Value': 'WebServerGroup'}]

        self.assertEqual(expected, self.group._tags())


class TestInitialGroupSize(common.HeatTestCase):
    scenarios = [
        ('000', dict(mins=0, maxs=0, desired=0, expected=0)),
        ('040', dict(mins=0, maxs=4, desired=0, expected=0)),
        ('253', dict(mins=2, maxs=5, desired=3, expected=3)),
        ('14n', dict(mins=1, maxs=4, desired=None, expected=1)),
    ]

    def test_initial_size(self):
        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['MinSize'] = self.mins
        properties['MaxSize'] = self.maxs
        properties['DesiredCapacity'] = self.desired
        stack = utils.parse_stack(t, params=inline_templates.as_params)
        group = stack['WebServerGroup']
        with mock.patch.object(group, '_create_template') as mock_cre_temp:
            group.child_template()
            mock_cre_temp.assert_called_once_with(self.expected)


class TestGroupAdjust(common.HeatTestCase):
    def setUp(self):
        super(TestGroupAdjust, self).setUp()

        t = template_format.parse(as_template)
        self.stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.group = self.stack['WebServerGroup']
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()
        self.stub_SnapshotConstraint_validate()
        self.assertIsNone(self.group.validate())

    def test_scaling_policy_cooldown_toosoon(self):
        dont_call = self.patchobject(self.group, 'resize')
        self.patchobject(self.group, '_check_scaling_allowed',
                         side_effect=resource.NoActionRequired)
        self.assertRaises(resource.NoActionRequired,
                          self.group.adjust, 1)
        self.assertEqual([], dont_call.call_args_list)

    def test_scaling_same_capacity(self):
        """Don't resize when capacity is the same."""
        self.patchobject(grouputils, 'get_size', return_value=3)
        resize = self.patchobject(self.group, 'resize')
        finished_scaling = self.patchobject(self.group, '_finished_scaling')
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.assertRaises(resource.NoActionRequired,
                          self.group.adjust, 3,
                          adjustment_type='ExactCapacity')
        expected_notifies = []
        self.assertEqual(expected_notifies, notify.call_args_list)
        self.assertEqual(0, resize.call_count)
        self.assertEqual(0, finished_scaling.call_count)

    def test_scaling_update_in_progress(self):
        """Don't resize when update in progress"""
        self.group.state_set('UPDATE', 'IN_PROGRESS')
        resize = self.patchobject(self.group, 'resize')
        finished_scaling = self.patchobject(self.group, '_finished_scaling')
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.assertRaises(resource.NoActionRequired,
                          self.group.adjust, 3,
                          adjustment_type='ExactCapacity')
        expected_notifies = []
        self.assertEqual(expected_notifies, notify.call_args_list)
        self.assertEqual(0, resize.call_count)
        self.assertEqual(0, finished_scaling.call_count)

    def test_scale_up_min_adjustment(self):
        self.patchobject(grouputils, 'get_size', return_value=1)
        resize = self.patchobject(self.group, 'resize')
        finished_scaling = self.patchobject(self.group, '_finished_scaling')
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.patchobject(self.group, '_check_scaling_allowed')
        self.group.adjust(33, adjustment_type='PercentChangeInCapacity',
                          min_adjustment_step=2)

        expected_notifies = [
            mock.call(
                capacity=1, suffix='start',
                adjustment_type='PercentChangeInCapacity',
                groupname='WebServerGroup',
                message='Start resizing the group WebServerGroup',
                adjustment=33,
                stack=self.group.stack),
            mock.call(
                capacity=3, suffix='end',
                adjustment_type='PercentChangeInCapacity',
                groupname='WebServerGroup',
                message='End resizing the group WebServerGroup',
                adjustment=33,
                stack=self.group.stack)]

        self.assertEqual(expected_notifies, notify.call_args_list)
        resize.assert_called_once_with(3)
        finished_scaling.assert_called_once_with(
            None,
            'PercentChangeInCapacity : 33',
            size_changed=True)

    def test_scale_down_min_adjustment(self):
        self.patchobject(grouputils, 'get_size', return_value=5)
        resize = self.patchobject(self.group, 'resize')
        finished_scaling = self.patchobject(self.group, '_finished_scaling')
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.patchobject(self.group, '_check_scaling_allowed')
        self.group.adjust(-33, adjustment_type='PercentChangeInCapacity',
                          min_adjustment_step=2)

        expected_notifies = [
            mock.call(
                capacity=5, suffix='start',
                adjustment_type='PercentChangeInCapacity',
                groupname='WebServerGroup',
                message='Start resizing the group WebServerGroup',
                adjustment=-33,
                stack=self.group.stack),
            mock.call(
                capacity=3, suffix='end',
                adjustment_type='PercentChangeInCapacity',
                groupname='WebServerGroup',
                message='End resizing the group WebServerGroup',
                adjustment=-33,
                stack=self.group.stack)]

        self.assertEqual(expected_notifies, notify.call_args_list)
        resize.assert_called_once_with(3)
        finished_scaling.assert_called_once_with(
            None,
            'PercentChangeInCapacity : -33',
            size_changed=True)

    def test_scaling_policy_cooldown_ok(self):
        self.patchobject(grouputils, 'get_size', return_value=0)
        resize = self.patchobject(self.group, 'resize')
        finished_scaling = self.patchobject(self.group, '_finished_scaling')
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.patchobject(self.group, '_check_scaling_allowed')
        self.group.adjust(1)

        expected_notifies = [
            mock.call(
                capacity=0, suffix='start', adjustment_type='ChangeInCapacity',
                groupname='WebServerGroup',
                message='Start resizing the group WebServerGroup',
                adjustment=1,
                stack=self.group.stack),
            mock.call(
                capacity=1, suffix='end',
                adjustment_type='ChangeInCapacity',
                groupname='WebServerGroup',
                message='End resizing the group WebServerGroup',
                adjustment=1,
                stack=self.group.stack)]

        self.assertEqual(expected_notifies, notify.call_args_list)
        resize.assert_called_once_with(1)
        finished_scaling.assert_called_once_with(None,
                                                 'ChangeInCapacity : 1',
                                                 size_changed=True)
        grouputils.get_size.assert_called_once_with(self.group)

    def test_scaling_policy_resize_fail(self):
        self.patchobject(grouputils, 'get_size', return_value=0)
        self.patchobject(self.group, 'resize',
                         side_effect=ValueError('test error'))
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.patchobject(self.group, '_check_scaling_allowed')
        self.patchobject(self.group, '_finished_scaling')
        self.assertRaises(ValueError, self.group.adjust, 1)

        expected_notifies = [
            mock.call(
                capacity=0, suffix='start',
                adjustment_type='ChangeInCapacity',
                groupname='WebServerGroup',
                message='Start resizing the group WebServerGroup',
                adjustment=1,
                stack=self.group.stack),
            mock.call(
                capacity=0, suffix='error',
                adjustment_type='ChangeInCapacity',
                groupname='WebServerGroup',
                message='test error',
                adjustment=1,
                stack=self.group.stack)]

        self.assertEqual(expected_notifies, notify.call_args_list)
        grouputils.get_size.assert_called_with(self.group)

    def test_notification_send_if_resize_failed(self):
        """If resize failed, the capacity of group might have been changed"""
        self.patchobject(grouputils, 'get_size', side_effect=[3, 4])
        self.patchobject(self.group, 'resize',
                         side_effect=ValueError('test error'))
        notify = self.patch('heat.engine.notification.autoscaling.send')
        self.patchobject(self.group, '_check_scaling_allowed')
        self.patchobject(self.group, '_finished_scaling')

        self.assertRaises(ValueError, self.group.adjust,
                          5, adjustment_type='ExactCapacity')

        expected_notifies = [
            mock.call(
                capacity=3, suffix='start',
                adjustment_type='ExactCapacity',
                groupname='WebServerGroup',
                message='Start resizing the group WebServerGroup',
                adjustment=5,
                stack=self.group.stack),
            mock.call(
                capacity=4, suffix='error',
                adjustment_type='ExactCapacity',
                groupname='WebServerGroup',
                message='test error',
                adjustment=5,
                stack=self.group.stack)]

        self.assertEqual(expected_notifies, notify.call_args_list)
        self.group.resize.assert_called_once_with(5)
        grouputils.get_size.assert_has_calls([mock.call(self.group),
                                              mock.call(self.group)])


class TestGroupCrud(common.HeatTestCase):
    def setUp(self):
        super(TestGroupCrud, self).setUp()
        t = template_format.parse(as_template)
        self.stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.group = self.stack['WebServerGroup']
        self.assertIsNone(self.group.validate())

    def test_handle_create(self):
        self.group.create_with_template = mock.Mock(return_value=None)
        self.group.child_template = mock.Mock(return_value='{}')

        self.group.handle_create()

        self.group.child_template.assert_called_once_with()
        self.group.create_with_template.assert_called_once_with('{}')

    def test_handle_update_desired_cap(self):
        self.group._try_rolling_update = mock.Mock(return_value=None)
        self.group.resize = mock.Mock(return_value=None)
        props = {'DesiredCapacity': 4,
                 'MinSize': 0,
                 'MaxSize': 6}
        self.group._get_new_capacity = mock.Mock(return_value=4)
        defn = rsrc_defn.ResourceDefinition(
            'nopayload',
            'AWS::AutoScaling::AutoScalingGroup',
            props)

        self.group.handle_update(defn, None, props)
        self.group.resize.assert_called_once_with(4)
        self.group._try_rolling_update.assert_called_once_with(props)

    def test_handle_update_desired_nocap(self):
        self.group._try_rolling_update = mock.Mock(return_value=None)
        self.group.resize = mock.Mock(return_value=None)
        get_size = self.patchobject(grouputils, 'get_size')
        get_size.return_value = 4

        props = {'MinSize': 0,
                 'MaxSize': 6}
        defn = rsrc_defn.ResourceDefinition(
            'nopayload',
            'AWS::AutoScaling::AutoScalingGroup',
            props)

        self.group.handle_update(defn, None, props)
        self.group.resize.assert_called_once_with(4)
        self.group._try_rolling_update.assert_called_once_with(props)

    def test_conf_properties_vpc_zone(self):
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()
        self.stub_SnapshotConstraint_validate()

        t = template_format.parse(as_template)
        properties = t['Resources']['WebServerGroup']['Properties']
        properties['VPCZoneIdentifier'] = ['xxxx']

        stack = utils.parse_stack(t, params=inline_templates.as_params)
        # create the launch configuration resource
        conf = stack['LaunchConfig']
        self.assertIsNone(conf.validate())
        scheduler.TaskRunner(conf.create)()
        self.assertEqual((conf.CREATE, conf.COMPLETE), conf.state)

        group = stack['WebServerGroup']
        config, props = group._get_conf_properties()
        self.assertEqual('xxxx', props['SubnetId'])
        conf.delete()

    def test_update_in_failed(self):
        self.group.state_set('CREATE', 'FAILED')
        # to update the failed asg
        self.group.resize = mock.Mock(return_value=None)

        new_defn = rsrc_defn.ResourceDefinition(
            'asg', 'AWS::AutoScaling::AutoScalingGroup',
            {'AvailabilityZones': ['nova'],
             'LaunchConfigurationName': 'config',
             'MaxSize': 5,
             'MinSize': 1,
             'DesiredCapacity': 2})

        self.group.handle_update(new_defn, None, None)
        self.group.resize.assert_called_once_with(2)


def asg_tmpl_with_bad_updt_policy():
    t = template_format.parse(inline_templates.as_template)
    ag = t['Resources']['WebServerGroup']
    ag["UpdatePolicy"] = {"foo": {}}
    return json.dumps(t)


def asg_tmpl_with_default_updt_policy():
    t = template_format.parse(inline_templates.as_template)
    ag = t['Resources']['WebServerGroup']
    ag["UpdatePolicy"] = {"AutoScalingRollingUpdate": {}}
    return json.dumps(t)


def asg_tmpl_with_updt_policy():
    t = template_format.parse(inline_templates.as_template)
    ag = t['Resources']['WebServerGroup']
    ag["UpdatePolicy"] = {"AutoScalingRollingUpdate": {
        "MinInstancesInService": "1",
        "MaxBatchSize": "2",
        "PauseTime": "PT1S"
    }}
    return json.dumps(t)


class RollingUpdatePolicyTest(common.HeatTestCase):
    def setUp(self):
        super(RollingUpdatePolicyTest, self).setUp()
        self.fc = fakes_nova.FakeClient()
        self.stub_keystoneclient(username='test_stack.CfnLBUser')
        self.stub_ImageConstraint_validate()
        self.stub_FlavorConstraint_validate()
        self.stub_SnapshotConstraint_validate()

    def test_parse_without_update_policy(self):
        tmpl = template_format.parse(inline_templates.as_template)
        stack = utils.parse_stack(tmpl, params=inline_templates.as_params)

        stack.validate()
        grp = stack['WebServerGroup']
        self.assertFalse(grp.update_policy['AutoScalingRollingUpdate'])

    def test_parse_with_update_policy(self):
        tmpl = template_format.parse(asg_tmpl_with_updt_policy())
        stack = utils.parse_stack(tmpl, params=inline_templates.as_params)

        stack.validate()
        tmpl_grp = tmpl['Resources']['WebServerGroup']
        tmpl_policy = tmpl_grp['UpdatePolicy']['AutoScalingRollingUpdate']
        tmpl_batch_sz = int(tmpl_policy['MaxBatchSize'])
        grp = stack['WebServerGroup']
        self.assertTrue(grp.update_policy)
        self.assertEqual(1, len(grp.update_policy))
        self.assertIn('AutoScalingRollingUpdate', grp.update_policy)
        policy = grp.update_policy['AutoScalingRollingUpdate']
        self.assertIsNotNone(policy)
        self.assertGreater(len(policy), 0)
        self.assertEqual(1, int(policy['MinInstancesInService']))
        self.assertEqual(tmpl_batch_sz, int(policy['MaxBatchSize']))
        self.assertEqual('PT1S', policy['PauseTime'])

    def test_parse_with_default_update_policy(self):
        tmpl = template_format.parse(asg_tmpl_with_default_updt_policy())
        stack = utils.parse_stack(tmpl, params=inline_templates.as_params)

        stack.validate()
        grp = stack['WebServerGroup']
        self.assertTrue(grp.update_policy)
        self.assertEqual(1, len(grp.update_policy))
        self.assertIn('AutoScalingRollingUpdate', grp.update_policy)
        policy = grp.update_policy['AutoScalingRollingUpdate']
        self.assertIsNotNone(policy)
        self.assertGreater(len(policy), 0)
        self.assertEqual(0, int(policy['MinInstancesInService']))
        self.assertEqual(1, int(policy['MaxBatchSize']))
        self.assertEqual('PT0S', policy['PauseTime'])

    def test_parse_with_bad_update_policy(self):
        tmpl = template_format.parse(asg_tmpl_with_bad_updt_policy())
        stack = utils.parse_stack(tmpl, params=inline_templates.as_params)
        error = self.assertRaises(
            exception.StackValidationFailed, stack.validate)
        self.assertIn("foo", str(error))

    def test_parse_with_bad_pausetime_in_update_policy(self):
        tmpl = template_format.parse(asg_tmpl_with_default_updt_policy())
        group = tmpl['Resources']['WebServerGroup']
        policy = group['UpdatePolicy']['AutoScalingRollingUpdate']
        policy['PauseTime'] = 'P1YT1H'
        stack = utils.parse_stack(tmpl, params=inline_templates.as_params)
        error = self.assertRaises(
            exception.StackValidationFailed, stack.validate)
        self.assertIn("Only ISO 8601 duration format", str(error))


class RollingUpdatePolicyDiffTest(common.HeatTestCase):
    def setUp(self):
        super(RollingUpdatePolicyDiffTest, self).setUp()
        self.fc = fakes_nova.FakeClient()
        self.stub_keystoneclient(username='test_stack.CfnLBUser')

    def validate_update_policy_diff(self, current, updated):
        # load current stack
        current_tmpl = template_format.parse(current)
        current_stack = utils.parse_stack(current_tmpl,
                                          params=inline_templates.as_params)

        # get the json snippet for the current InstanceGroup resource
        current_grp = current_stack['WebServerGroup']
        current_snippets = dict((n, r.frozen_definition())
                                for n, r in current_stack.items())
        current_grp_json = current_snippets[current_grp.name]

        # load the updated stack
        updated_tmpl = template_format.parse(updated)
        updated_stack = utils.parse_stack(updated_tmpl,
                                          params=inline_templates.as_params)

        # get the updated json snippet for the InstanceGroup resource in the
        # context of the current stack
        updated_grp = updated_stack['WebServerGroup']
        updated_grp_json = updated_grp.t.freeze()

        # identify the template difference
        tmpl_diff = updated_grp.update_template_diff(
            updated_grp_json, current_grp_json)
        self.assertTrue(tmpl_diff.update_policy_changed())

        # test application of the new update policy in handle_update
        current_grp._try_rolling_update = mock.MagicMock()
        current_grp.resize = mock.MagicMock()
        current_grp.handle_update(updated_grp_json, tmpl_diff, None)
        self.assertEqual(updated_grp_json._update_policy or {},
                         current_grp.update_policy.data)

    def test_update_policy_added(self):
        self.validate_update_policy_diff(inline_templates.as_template,
                                         asg_tmpl_with_updt_policy())

    def test_update_policy_updated(self):
        updt_template = json.loads(asg_tmpl_with_updt_policy())
        grp = updt_template['Resources']['WebServerGroup']
        policy = grp['UpdatePolicy']['AutoScalingRollingUpdate']
        policy['MinInstancesInService'] = '2'
        policy['MaxBatchSize'] = '4'
        policy['PauseTime'] = 'PT1M30S'
        self.validate_update_policy_diff(asg_tmpl_with_updt_policy(),
                                         json.dumps(updt_template))

    def test_update_policy_removed(self):
        self.validate_update_policy_diff(asg_tmpl_with_updt_policy(),
                                         inline_templates.as_template)


class TestCooldownMixin(common.HeatTestCase):
    def setUp(self):
        super(TestCooldownMixin, self).setUp()
        t = template_format.parse(inline_templates.as_template)
        self.stack = utils.parse_stack(t, params=inline_templates.as_params)
        self.stack.store()
        self.group = self.stack['WebServerGroup']
        self.group.state_set('CREATE', 'COMPLETE')

    def test_cooldown_is_in_progress_toosoon(self):
        cooldown_end = timeutils.utcnow() + datetime.timedelta(seconds=60)
        previous_meta = {'cooldown_end': {
            cooldown_end.isoformat(): 'ChangeInCapacity : 1'}}
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertRaises(resource.NoActionRequired,
                          self.group._check_scaling_allowed,
                          60)

    def test_cooldown_is_in_progress_toosoon_legacy(self):
        now = timeutils.utcnow()
        previous_meta = {'cooldown': {
            now.isoformat(): 'ChangeInCapacity : 1'}}
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertRaises(resource.NoActionRequired,
                          self.group._check_scaling_allowed,
                          60)

    def test_cooldown_is_in_progress_scaling_unfinished(self):
        previous_meta = {'scaling_in_progress': True}
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertRaises(resource.NoActionRequired,
                          self.group._check_scaling_allowed,
                          60)

    def test_scaling_not_in_progress_legacy(self):
        awhile_ago = timeutils.utcnow() - datetime.timedelta(seconds=100)
        previous_meta = {
            'cooldown': {
                awhile_ago.isoformat(): 'ChangeInCapacity : 1'
            },
            'scaling_in_progress': False
        }
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertIsNone(self.group._check_scaling_allowed(60))

    def test_scaling_not_in_progress(self):
        awhile_after = timeutils.utcnow() + datetime.timedelta(seconds=60)
        previous_meta = {
            'cooldown_end': {
                awhile_after.isoformat(): 'ChangeInCapacity : 1'
            },
            'scaling_in_progress': False
        }
        timeutils.set_time_override()
        timeutils.advance_time_seconds(100)
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertIsNone(self.group._check_scaling_allowed(60))
        timeutils.clear_time_override()

    def test_scaling_policy_cooldown_zero(self):
        now = timeutils.utcnow()
        previous_meta = {
            'cooldown_end': {
                now.isoformat(): 'ChangeInCapacity : 1'
            },
            'scaling_in_progress': False
        }
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertIsNone(self.group._check_scaling_allowed(60))

    def test_scaling_policy_cooldown_none(self):
        now = timeutils.utcnow()
        previous_meta = {
            'cooldown_end': {
                now.isoformat(): 'ChangeInCapacity : 1'
            },
            'scaling_in_progress': False
        }
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        self.assertIsNone(self.group._check_scaling_allowed(None))

    def test_metadata_is_written(self):
        nowish = timeutils.utcnow()
        reason = 'cool as'
        meta_set = self.patchobject(self.group, 'metadata_set')
        self.patchobject(timeutils, 'utcnow', return_value=nowish)
        self.group._finished_scaling(60, reason)
        cooldown_end = nowish + datetime.timedelta(seconds=60)
        meta_set.assert_called_once_with(
            {'cooldown_end': {cooldown_end.isoformat(): reason},
             'scaling_in_progress': False})

    def test_metadata_is_written_update(self):
        nowish = timeutils.utcnow()
        reason = 'cool as'
        prev_cooldown_end = nowish + datetime.timedelta(seconds=100)
        previous_meta = {
            'cooldown_end': {
                prev_cooldown_end.isoformat(): 'ChangeInCapacity : 1'
            }
        }
        self.patchobject(self.group, 'metadata_get',
                         return_value=previous_meta)
        meta_set = self.patchobject(self.group, 'metadata_set')
        self.patchobject(timeutils, 'utcnow', return_value=nowish)
        self.group._finished_scaling(60, reason)
        meta_set.assert_called_once_with(
            {'cooldown_end': {prev_cooldown_end.isoformat(): reason},
             'scaling_in_progress': False})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os
import subprocess
from unittest import mock


from heat.cloudinit import loguserdata
from heat.tests import common


class LoguserdataTest(common.HeatTestCase):

    @mock.patch('subprocess.Popen')
    def test_call(self, mock_popen):
        # Setup
        mock_popen.return_value = mock.MagicMock()
        mock_popen.return_value.communicate.return_value = ['a', 'b']
        mock_popen.return_value.returncode = 0

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        mock_popen.assert_called_once_with(['foo', 'bar'],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
        self.assertEqual(0, return_code)

    @mock.patch('sys.exc_info')
    @mock.patch('subprocess.Popen')
    def test_call_oserror_enoexec(self, mock_popen, mock_exc_info):
        # Setup
        mock_popen.side_effect = OSError()
        no_exec = mock.MagicMock(errno=errno.ENOEXEC)
        mock_exc_info.return_value = None, no_exec, None

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        self.assertEqual(os.EX_OK, return_code)

    @mock.patch('sys.exc_info')
    @mock.patch('subprocess.Popen')
    def test_call_oserror_other(self, mock_popen, mock_exc_info):
        # Setup
        mock_popen.side_effect = OSError()
        no_exec = mock.MagicMock(errno='foo')
        mock_exc_info.return_value = None, no_exec, None

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        self.assertEqual(os.EX_OSERR, return_code)

    @mock.patch('sys.exc_info')
    @mock.patch('subprocess.Popen')
    def test_call_exception(self, mock_popen, mock_exc_info):
        # Setup
        mock_popen.side_effect = Exception()
        no_exec = mock.MagicMock(errno='irrelevant')
        mock_exc_info.return_value = None, no_exec, None

        # Test
        return_code = loguserdata.call(['foo', 'bar'])

        # Verify
        self.assertEqual(os.EX_SOFTWARE, return_code)

    @mock.patch('os.chmod')
    @mock.patch('heat.cloudinit.loguserdata.call')
    def test_main(self, mock_call, mock_chmod):
        # Setup
        mock_call.return_value = 10

        # Test
        return_code = loguserdata.main()

        # Verify
        expected_path = os.path.join(loguserdata.VAR_PATH, 'cfn-userdata')
        mock_chmod.assert_called_once_with(expected_path, int('700', 8))
        self.assertEqual(10, return_code)



#
#    Copyright 2015 IBM Corp.
#
#    All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support


class Listener(none_resource.NoneResource):
    """A resource for managing LBaaS v2 Listeners.

    This resource creates and manages Neutron LBaaS v2 Listeners,
    which represent a listening endpoint for the vip.
    """

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='21.0.0',
        message=_('Use octavia instead.'),
        previous_status=support.SupportStatus(version='6.0.0')
    )


def resource_mapping():
    return {
        'OS::Neutron::LBaaS::Listener': Listener,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
import swiftclient.client as sc

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.aws.s3 import s3
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


swift_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "Template to test S3 Bucket resources",
  "Resources" : {
    "S3BucketWebsite" : {
      "Type" : "AWS::S3::Bucket",
      "DeletionPolicy" : "Delete",
      "Properties" : {
        "AccessControl" : "PublicRead",
        "WebsiteConfiguration" : {
          "IndexDocument" : "index.html",
          "ErrorDocument" : "error.html"
         }
      }
    },
    "SwiftContainer": {
         "Type": "OS::Swift::Container",
         "Properties": {
            "S3Bucket": {"Ref" : "S3Bucket"},
         }
      },
    "S3Bucket" : {
      "Type" : "AWS::S3::Bucket",
      "Properties" : {
        "AccessControl" : "Private"
      }
    },
    "S3Bucket_with_tags" : {
      "Type" : "AWS::S3::Bucket",
      "Properties" : {
        "Tags" : [{"Key": "greeting", "Value": "hello"},
                  {"Key": "location", "Value": "here"}]
      }
    }
  }
}
'''


class s3Test(common.HeatTestCase):
    def setUp(self):
        super(s3Test, self).setUp()
        self.mock_con = mock.Mock(spec=sc.Connection)
        self.patchobject(s3.S3Bucket, 'client',
                         return_value=self.mock_con)

    def create_resource(self, t, stack, resource_name):
        resource_defns = stack.t.resource_definitions(stack)
        rsrc = s3.S3Bucket('test_resource',
                           resource_defns[resource_name],
                           stack)
        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)
        return rsrc

    def test_attributes(self):
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.get_auth.return_value = (
            'http://server.test:8080/v_2', None)
        self.mock_con.delete_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3Bucket')

        ref_id = rsrc.FnGetRefId()
        self.assertEqual(container_name, ref_id)

        self.assertEqual('server.test', rsrc.FnGetAtt('DomainName'))
        url = 'http://server.test:8080/v_2/%s' % ref_id

        self.assertEqual(url, rsrc.FnGetAtt('WebsiteURL'))

        self.assertRaises(exception.InvalidTemplateAttribute,
                          rsrc.FnGetAtt, 'Foo')

        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username'}
        )
        self.mock_con.get_auth.assert_called_with()
        self.assertEqual(2, self.mock_con.get_auth.call_count)
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_public_read(self):
        t = template_format.parse(swift_template)
        properties = t['Resources']['S3Bucket']['Properties']
        properties['AccessControl'] = 'PublicRead'
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            utils.PhysName(stack.name, 'test_resource'),
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': '.r:*'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_tags(self):
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3Bucket_with_tags')
        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            utils.PhysName(stack.name, 'test_resource'),
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username',
             'X-Container-Meta-S3-Tag-greeting': 'hello',
             'X-Container-Meta-S3-Tag-location': 'here'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_public_read_write(self):
        t = template_format.parse(swift_template)
        properties = t['Resources']['S3Bucket']['Properties']
        properties['AccessControl'] = 'PublicReadWrite'
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': '.r:*',
             'X-Container-Read': '.r:*'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_authenticated_read(self):
        t = template_format.parse(swift_template)
        properties = t['Resources']['S3Bucket']['Properties']
        properties['AccessControl'] = 'AuthenticatedRead'
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_website(self):
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3BucketWebsite')
        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Meta-Web-Error': 'error.html',
             'X-Container-Meta-Web-Index': 'index.html',
             'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': '.r:*'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_delete_exception(self):
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.side_effect = sc.ClientException(
            'Test Delete Failure')

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        self.assertRaises(exception.ResourceFailure,
                          scheduler.TaskRunner(rsrc.delete))
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_delete_not_found(self):
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.side_effect = sc.ClientException(
            'Gone', http_status=404)

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        scheduler.TaskRunner(rsrc.delete)()
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username'})
        self.mock_con.delete_container.assert_called_once_with(container_name)

    def test_delete_conflict_not_empty(self):
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.side_effect = sc.ClientException(
            'Not empty', http_status=409)
        self.mock_con.get_container.return_value = (
            {'name': container_name}, [{'name': 'test_object'}])

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        deleter = scheduler.TaskRunner(rsrc.delete)
        ex = self.assertRaises(exception.ResourceFailure, deleter)
        self.assertIn("ResourceActionNotSupported: resources.test_resource: "
                      "The bucket you tried to delete is not empty",
                      str(ex))
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username'})
        self.mock_con.delete_container.assert_called_once_with(container_name)
        self.mock_con.get_container.assert_called_once_with(container_name)

    def test_delete_conflict_empty(self):
        cfg.CONF.set_override('action_retry_limit', 0)
        t = template_format.parse(swift_template)
        stack = utils.parse_stack(t)

        container_name = utils.PhysName(stack.name, 'test_resource')
        self.mock_con.put_container.return_value = None
        self.mock_con.delete_container.side_effect = sc.ClientException(
            'Conflict', http_status=409)
        self.mock_con.get_container.return_value = (
            {'name': container_name}, [])

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        deleter = scheduler.TaskRunner(rsrc.delete)
        ex = self.assertRaises(exception.ResourceFailure, deleter)
        self.assertIn("Conflict", str(ex))
        self.mock_con.put_container.assert_called_once_with(
            container_name,
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username'})
        self.mock_con.delete_container.assert_called_once_with(container_name)
        self.mock_con.get_container.assert_called_once_with(container_name)

    def test_delete_retain(self):
        t = template_format.parse(swift_template)
        bucket = t['Resources']['S3Bucket']
        bucket['DeletionPolicy'] = 'Retain'
        stack = utils.parse_stack(t)

        # first run, with retain policy
        self.mock_con.put_container.return_value = None

        rsrc = self.create_resource(t, stack, 'S3Bucket')
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)
        self.mock_con.put_container.assert_called_once_with(
            utils.PhysName(stack.name, 'test_resource'),
            {'X-Container-Write': 'test_tenant:test_username',
             'X-Container-Read': 'test_tenant:test_username'})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine import resource
from heat.engine.resources.openstack.manila import security_service
from heat.engine import scheduler
from heat.engine import template
from heat.tests import common
from heat.tests import utils

stack_template = '''
heat_template_version: 2013-05-23

resources:
  security_service:
    type: OS::Manila::SecurityService
    properties:
      name: my_security_service
      domain: test-domain
      dns_ip: 1.1.1.1
      type: ldap
      server: test-server
      user: test-user
      password: test-password
'''

stack_template_update = '''
heat_template_version: 2013-05-23

resources:
  security_service:
    type: OS::Manila::SecurityService
    properties:
      name: fake_security_service
      domain: fake-domain
      dns_ip: 1.1.1.1
      type: ldap
      server: fake-server
'''

stack_template_update_replace = '''
heat_template_version: 2013-05-23

resources:
  security_service:
    type: OS::Manila::SecurityService
    properties:
      name: my_security_service
      domain: test-domain
      dns_ip: 1.1.1.1
      type: kerberos
      server: test-server
      user: test-user
      password: test-password
'''


class ManilaSecurityServiceTest(common.HeatTestCase):

    def setUp(self):
        super(ManilaSecurityServiceTest, self).setUp()

        t = template_format.parse(stack_template)
        self.stack = utils.parse_stack(t)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns['security_service']

        self.client = mock.Mock()
        self.patchobject(security_service.SecurityService, 'client',
                         return_value=self.client)

    def _create_resource(self, name, snippet, stack):
        ss = security_service.SecurityService(name, snippet, stack)
        value = mock.MagicMock(id='12345')
        self.client.security_services.create.return_value = value
        self.client.security_services.get.return_value = value
        scheduler.TaskRunner(ss.create)()
        args = self.client.security_services.create.call_args[1]
        self.assertEqual(self.rsrc_defn._properties, args)
        self.assertEqual('12345', ss.resource_id)
        return ss

    def test_create(self):
        ct = self._create_resource('security_service', self.rsrc_defn,
                                   self.stack)
        expected_state = (ct.CREATE, ct.COMPLETE)
        self.assertEqual(expected_state, ct.state)
        self.assertEqual('security_services', ct.entity)

    def test_create_failed(self):
        ss = security_service.SecurityService('security_service',
                                              self.rsrc_defn, self.stack)
        self.client.security_services.create.side_effect = Exception('error')

        exc = self.assertRaises(exception.ResourceFailure,
                                scheduler.TaskRunner(ss.create))
        expected_state = (ss.CREATE, ss.FAILED)
        self.assertEqual(expected_state, ss.state)
        self.assertIn('Exception: resources.security_service: error',
                      str(exc))

    def test_update(self):
        ss = self._create_resource('security_service', self.rsrc_defn,
                                   self.stack)
        t = template_format.parse(stack_template_update)
        rsrc_defns = template.Template(t).resource_definitions(self.stack)
        new_ss = rsrc_defns['security_service']
        scheduler.TaskRunner(ss.update, new_ss)()
        args = {
            'domain': 'fake-domain',
            'password': None,
            'user': None,
            'server': 'fake-server',
            'name': 'fake_security_service'
        }
        self.client.security_services.update.assert_called_once_with(
            '12345', **args)
        self.assertEqual((ss.UPDATE, ss.COMPLETE), ss.state)

    def test_update_replace(self):
        ss = self._create_resource('security_service', self.rsrc_defn,
                                   self.stack)
        t = template_format.parse(stack_template_update_replace)
        rsrc_defns = template.Template(t).resource_definitions(self.stack)
        new_ss = rsrc_defns['security_service']
        self.assertEqual(0, self.client.security_services.update.call_count)
        err = self.assertRaises(resource.UpdateReplace,
                                scheduler.TaskRunner(ss.update, new_ss))
        msg = 'The Resource security_service requires replacement.'
        self.assertEqual(msg, str(err))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for database migrations."""

from alembic import command as alembic_api
from alembic import script as alembic_script
import fixtures
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslotest import base as test_base
import sqlalchemy
import testtools

from heat.db import migration
from heat.db import models


class DBNotAllowed(Exception):
    pass


class BannedDBSchemaOperations(fixtures.Fixture):
    """Ban some operations for migrations"""
    def __init__(self, banned_resources=None):
        super(BannedDBSchemaOperations, self).__init__()
        self._banned_resources = banned_resources or []

    @staticmethod
    def _explode(resource, op):
        print('%s.%s()' % (resource, op))
        raise DBNotAllowed(
            'Operation %s.%s() is not allowed in a database migration' % (
                resource, op))

    def setUp(self):
        super(BannedDBSchemaOperations, self).setUp()
        for thing in self._banned_resources:
            self.useFixture(fixtures.MonkeyPatch(
                'sqlalchemy.%s.drop' % thing,
                lambda *a, **k: self._explode(thing, 'drop')))
            self.useFixture(fixtures.MonkeyPatch(
                'sqlalchemy.%s.alter' % thing,
                lambda *a, **k: self._explode(thing, 'alter')))


class TestBannedDBSchemaOperations(testtools.TestCase):
    def test_column(self):
        column = sqlalchemy.Column()
        with BannedDBSchemaOperations(['Column']):
            self.assertRaises(DBNotAllowed, column.drop)
            self.assertRaises(DBNotAllowed, column.alter)

    def test_table(self):
        table = sqlalchemy.Table("foo", sqlalchemy.MetaData())
        with BannedDBSchemaOperations(['Table']):
            self.assertRaises(DBNotAllowed, table.drop)
            self.assertRaises(DBNotAllowed, table.alter)


class HeatModelsMigrationsSync(test_migrations.ModelsMigrationsSync):

    def setUp(self):
        super().setUp()

        self.engine = enginefacade.writer.get_engine()
        self.sessionmaker = enginefacade.writer.get_sessionmaker()

    def get_metadata(self):
        return models.BASE.metadata

    def get_engine(self):
        return self.engine

    def db_sync(self, engine):
        migration.db_sync(engine=engine)

    def include_object(self, object_, name, type_, reflected, compare_to):
        return True


class ModelsMigrationsSyncMysql(
    HeatModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class ModelsMigrationsSyncPostgres(
    HeatModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class ModelsMigrationsSyncSQLite(
    HeatModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    pass


class DatabaseSanityChecks(
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        # self.patch(api, 'get_engine', lambda: self.engine)
        self.config = migration._find_alembic_conf()
        self.init_version = migration.ALEMBIC_INIT_VERSION

    def test_single_base_revision(self):
        """Ensure we only have a single base revision.

        There's no good reason for us to have diverging history, so validate
        that only one base revision exists. This will prevent simple errors
        where people forget to specify the base revision. If this fail for your
        change, look for migrations that do not have a 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_bases()))

    def test_single_head_revision(self):
        """Ensure we only have a single head revision.

        There's no good reason for us to have diverging history, so validate
        that only one head revision exists. This will prevent merge conflicts
        adding additional head revision points. If this fail for your change,
        look for migrations with the same 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_heads()))


class MigrationsWalk(
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4

    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        # self.patch(api, 'get_engine', lambda: self.engine)
        self.config = migration._find_alembic_conf()
        self.init_version = migration.ALEMBIC_INIT_VERSION

    def _migrate_up(self, revision, connection):
        check_method = getattr(self, f'_check_{revision}', None)
        if revision != self.init_version:  # no tests for the initial revision
            self.assertIsNotNone(
                check_method,
                f"DB Migration {revision} doesn't have a test; add one"
            )

        pre_upgrade = getattr(self, f'_pre_upgrade_{revision}', None)
        if pre_upgrade:
            pre_upgrade(connection)

        alembic_api.upgrade(self.config, revision)

        if check_method:
            check_method(connection)

    def test_walk_versions(self):
        with self.engine.begin() as connection:
            self.config.attributes['connection'] = connection
            script = alembic_script.ScriptDirectory.from_config(self.config)
            revisions = list(script.walk_revisions())
            # Need revisions from older to newer so the walk works as intended
            revisions.reverse()
            for revision_script in revisions:
                self._migrate_up(revision_script.revision, connection)


class TestMigrationsWalkSQLite(
    MigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    pass


class TestMigrationsWalkMySQL(
    MigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestMigrationsWalkPostgreSQL(
    MigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.octavia import octavia_base
from heat.engine import translation


class HealthMonitor(octavia_base.OctaviaBase):
    """A resource to handle load balancer health monitors.

    This resource creates and manages octavia healthmonitors,
    which watches status of the load balanced servers.
    """

    # Properties inputs for the resources create/update.
    PROPERTIES = (
        ADMIN_STATE_UP, DELAY, EXPECTED_CODES, HTTP_METHOD,
        MAX_RETRIES, POOL, TIMEOUT, TYPE, URL_PATH, TENANT_ID
    ) = (
        'admin_state_up', 'delay', 'expected_codes', 'http_method',
        'max_retries', 'pool', 'timeout', 'type', 'url_path', 'tenant_id'
    )

    # Supported HTTP methods
    HTTP_METHODS = (
        GET, HEAT, POST, PUT, DELETE, TRACE, OPTIONS,
        CONNECT, PATCH
    ) = (
        'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'TRACE', 'OPTIONS',
        'CONNECT', 'PATCH'
    )

    # Supported output attributes of the resources.
    ATTRIBUTES = (POOLS_ATTR) = ('pools')

    properties_schema = {
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('The administrative state of the health monitor.'),
            default=True,
            update_allowed=True
        ),
        DELAY: properties.Schema(
            properties.Schema.INTEGER,
            _('The minimum time in seconds between regular connections '
              'of the member.'),
            required=True,
            update_allowed=True,
            constraints=[constraints.Range(min=0)]
        ),
        EXPECTED_CODES: properties.Schema(
            properties.Schema.STRING,
            _('The HTTP status codes expected in response from the '
              'member to declare it healthy. Specify one of the following '
              'values: a single value, such as 200. a list, such as 200, 202. '
              'a range, such as 200-204.'),
            update_allowed=True,
        ),
        HTTP_METHOD: properties.Schema(
            properties.Schema.STRING,
            _('The HTTP method used for requests by the monitor of type '
              'HTTP.'),
            update_allowed=True,
            constraints=[constraints.AllowedValues(HTTP_METHODS)]
        ),
        MAX_RETRIES: properties.Schema(
            properties.Schema.INTEGER,
            _('Number of permissible connection failures before changing the '
              'member status to INACTIVE.'),
            required=True,
            update_allowed=True,
            constraints=[constraints.Range(min=1, max=10)],
        ),
        POOL: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the load balancing pool.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('octavia.pool')
            ]
        ),
        TIMEOUT: properties.Schema(
            properties.Schema.INTEGER,
            _('Maximum number of seconds for a monitor to wait for a '
              'connection to be established before it times out.'),
            required=True,
            update_allowed=True,
            constraints=[constraints.Range(min=0)]
        ),
        TYPE: properties.Schema(
            properties.Schema.STRING,
            _('One of predefined health monitor types.'),
            required=True,
            constraints=[
                constraints.AllowedValues(['PING', 'TCP', 'HTTP', 'HTTPS',
                                           'UDP-CONNECT']),
            ]
        ),
        URL_PATH: properties.Schema(
            properties.Schema.STRING,
            _('The HTTP path used in the HTTP request used by the monitor to '
              'test a member health. A valid value is a string the begins '
              'with a forward slash (/).'),
            update_allowed=True,
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('ID of the tenant who owns the health monitor.')
        )
    }

    attributes_schema = {
        POOLS_ATTR: attributes.Schema(
            _('The list of Pools related to this monitor.'),
            type=attributes.Schema.LIST
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.POOL],
                client_plugin=self.client_plugin(),
                finder='get_pool',
            ),
        ]

    def _prepare_args(self, properties):
        props = dict((k, v) for k, v in properties.items() if v is not None)
        if self.POOL in props:
            props['pool_id'] = props.pop(self.POOL)
        return props

    def _resource_create(self, properties):
        return self.client().health_monitor_create(
            json={'healthmonitor': properties})['healthmonitor']

    def _resource_update(self, prop_diff):
        self.client().health_monitor_set(
            self.resource_id, json={'healthmonitor': prop_diff})

    def _resource_delete(self):
        self.client().health_monitor_delete(self.resource_id)

    def _show_resource(self):
        return self.client().health_monitor_show(self.resource_id)


def resource_mapping():
    return {
        'OS::Octavia::HealthMonitor': HealthMonitor,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.tests import common

from heat.scaling import rolling_update


class TestNeedsUpdate(common.HeatTestCase):
    scenarios = [
        ('4_4_0', dict(targ=4, curr=4, updated=0, result=True)),
        ('4_4_1', dict(targ=4, curr=4, updated=1, result=True)),
        ('4_4_3', dict(targ=4, curr=4, updated=3, result=True)),
        ('4_4_4', dict(targ=4, curr=4, updated=4, result=False)),
        ('4_4_5', dict(targ=4, curr=4, updated=5, result=False)),
        ('4_5_0', dict(targ=4, curr=5, updated=0, result=True)),
        ('4_5_1', dict(targ=4, curr=5, updated=1, result=True)),
        ('4_5_3', dict(targ=4, curr=5, updated=3, result=True)),
        ('4_5_4', dict(targ=4, curr=5, updated=4, result=True)),
        ('4_5_5', dict(targ=4, curr=5, updated=5, result=True)),
        ('4_3_0', dict(targ=4, curr=3, updated=0, result=True)),
        ('4_3_1', dict(targ=4, curr=3, updated=1, result=True)),
        ('4_3_2', dict(targ=4, curr=3, updated=2, result=True)),
        ('4_3_3', dict(targ=4, curr=3, updated=3, result=True)),
        ('4_3_4', dict(targ=4, curr=3, updated=4, result=True)),
    ]

    def test_needs_update(self):
        needs_update = rolling_update.needs_update(self.targ, self.curr,
                                                   self.updated)
        self.assertEqual(self.result, needs_update)


class TestNextBatch(common.HeatTestCase):

    scenarios = [
        ('4_4_0_1_0', dict(targ=4, curr=4, updated=0, bat_size=1, min_srv=0,
                           batch=(4, 1))),
        ('4_4_3_1_0', dict(targ=4, curr=4, updated=3, bat_size=1, min_srv=0,
                           batch=(4, 1))),
        ('4_4_0_1_4', dict(targ=4, curr=4, updated=0, bat_size=1, min_srv=4,
                           batch=(5, 1))),
        ('4_5_3_1_4', dict(targ=4, curr=5, updated=3, bat_size=1, min_srv=4,
                           batch=(5, 1))),
        ('4_5_4_1_4', dict(targ=4, curr=5, updated=4, bat_size=1, min_srv=4,
                           batch=(4, 0))),
        ('4_4_0_1_5', dict(targ=4, curr=4, updated=0, bat_size=1, min_srv=5,
                           batch=(5, 1))),
        ('4_5_3_1_5', dict(targ=4, curr=5, updated=3, bat_size=1, min_srv=5,
                           batch=(5, 1))),
        ('4_5_0_1_4', dict(targ=4, curr=5, updated=0, bat_size=1, min_srv=4,
                           batch=(5, 1))),
        ('4_5_1_1_4', dict(targ=4, curr=5, updated=1, bat_size=1, min_srv=4,
                           batch=(5, 1))),
        ('4_5_4_1_5', dict(targ=4, curr=5, updated=4, bat_size=1, min_srv=5,
                           batch=(4, 0))),
        ('4_4_0_2_0', dict(targ=4, curr=4, updated=0, bat_size=2, min_srv=0,
                           batch=(4, 2))),
        ('4_4_2_2_0', dict(targ=4, curr=4, updated=2, bat_size=2, min_srv=0,
                           batch=(4, 2))),
        ('4_4_0_2_4', dict(targ=4, curr=4, updated=0, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('4_6_2_2_4', dict(targ=4, curr=4, updated=0, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('4_6_4_2_4', dict(targ=4, curr=6, updated=4, bat_size=2, min_srv=4,
                           batch=(4, 0))),
        ('5_5_0_2_0', dict(targ=5, curr=5, updated=0, bat_size=2, min_srv=0,
                           batch=(5, 2))),
        ('5_5_4_2_0', dict(targ=5, curr=5, updated=4, bat_size=2, min_srv=0,
                           batch=(5, 1))),
        ('5_5_0_2_4', dict(targ=5, curr=5, updated=0, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('5_6_2_2_4', dict(targ=5, curr=6, updated=2, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('5_6_4_2_4', dict(targ=5, curr=6, updated=4, bat_size=2, min_srv=4,
                           batch=(5, 1))),
        ('5_7_2_2_4', dict(targ=5, curr=7, updated=2, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('3_3_0_2_0', dict(targ=3, curr=3, updated=0, bat_size=2, min_srv=0,
                           batch=(3, 2))),
        ('3_3_2_2_0', dict(targ=3, curr=3, updated=2, bat_size=2, min_srv=0,
                           batch=(3, 1))),
        ('3_3_0_2_4', dict(targ=3, curr=3, updated=0, bat_size=2, min_srv=4,
                           batch=(5, 2))),
        ('3_5_2_2_4', dict(targ=3, curr=5, updated=2, bat_size=2, min_srv=4,
                           batch=(4, 1))),
        ('3_5_3_2_4', dict(targ=3, curr=5, updated=3, bat_size=2, min_srv=4,
                           batch=(3, 0))),
        ('4_4_0_4_0', dict(targ=4, curr=4, updated=0, bat_size=4, min_srv=0,
                           batch=(4, 4))),
        ('4_4_0_5_0', dict(targ=4, curr=4, updated=0, bat_size=5, min_srv=0,
                           batch=(4, 4))),
        ('4_4_0_4_1', dict(targ=4, curr=4, updated=0, bat_size=4, min_srv=1,
                           batch=(5, 4))),
        ('4_4_4_4_1', dict(targ=4, curr=4, updated=4, bat_size=4, min_srv=1,
                           batch=(4, 0))),
        ('4_4_0_6_1', dict(targ=4, curr=4, updated=0, bat_size=6, min_srv=1,
                           batch=(5, 4))),
        ('4_4_4_6_1', dict(targ=4, curr=4, updated=4, bat_size=6, min_srv=1,
                           batch=(4, 0))),
        ('4_4_0_4_2', dict(targ=4, curr=4, updated=0, bat_size=4, min_srv=2,
                           batch=(6, 4))),
        ('4_4_4_4_2', dict(targ=4, curr=4, updated=4, bat_size=4, min_srv=2,
                           batch=(4, 0))),
        ('4_4_0_4_4', dict(targ=4, curr=4, updated=0, bat_size=4, min_srv=4,
                           batch=(8, 4))),
        ('4_4_4_4_4', dict(targ=4, curr=4, updated=4, bat_size=4, min_srv=4,
                           batch=(4, 0))),
        ('4_4_0_5_6', dict(targ=4, curr=4, updated=0, bat_size=5, min_srv=6,
                           batch=(8, 4))),
        ('4_4_4_5_6', dict(targ=4, curr=4, updated=4, bat_size=5, min_srv=6,
                           batch=(4, 0))),
        ('6_4_0_1_0', dict(targ=6, curr=4, updated=0, bat_size=1, min_srv=0,
                           batch=(5, 1))),
        ('6_5_1_1_0', dict(targ=6, curr=5, updated=1, bat_size=1, min_srv=0,
                           batch=(6, 1))),
        ('6_4_0_1_4', dict(targ=6, curr=4, updated=0, bat_size=1, min_srv=4,
                           batch=(5, 1))),
        ('6_5_1_1_4', dict(targ=6, curr=5, updated=1, bat_size=1, min_srv=4,
                           batch=(6, 1))),
        ('6_4_0_2_0', dict(targ=6, curr=4, updated=0, bat_size=2, min_srv=0,
                           batch=(6, 2))),
        ('6_4_0_2_4', dict(targ=6, curr=4, updated=0, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('6_5_0_2_0', dict(targ=6, curr=5, updated=0, bat_size=2, min_srv=0,
                           batch=(6, 2))),
        ('6_5_0_2_4', dict(targ=6, curr=5, updated=0, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('6_3_0_2_0', dict(targ=6, curr=3, updated=0, bat_size=2, min_srv=0,
                           batch=(5, 2))),
        ('6_5_2_2_0', dict(targ=6, curr=5, updated=2, bat_size=2, min_srv=0,
                           batch=(6, 2))),
        ('6_5_4_2_0', dict(targ=6, curr=5, updated=4, bat_size=2, min_srv=0,
                           batch=(6, 2))),
        ('6_3_0_2_4', dict(targ=6, curr=3, updated=0, bat_size=2, min_srv=4,
                           batch=(5, 2))),
        ('6_5_2_2_4', dict(targ=6, curr=5, updated=2, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('6_5_4_2_4', dict(targ=6, curr=5, updated=4, bat_size=2, min_srv=4,
                           batch=(6, 2))),
        ('6_4_0_4_0', dict(targ=6, curr=4, updated=0, bat_size=4, min_srv=0,
                           batch=(6, 4))),
        ('6_4_4_4_0', dict(targ=6, curr=4, updated=4, bat_size=4, min_srv=0,
                           batch=(6, 2))),
        ('6_4_0_5_0', dict(targ=6, curr=4, updated=0, bat_size=5, min_srv=0,
                           batch=(6, 5))),
        ('6_6_5_4_0', dict(targ=6, curr=6, updated=5, bat_size=4, min_srv=0,
                           batch=(6, 1))),
        ('6_4_0_4_1', dict(targ=6, curr=4, updated=0, bat_size=4, min_srv=1,
                           batch=(6, 4))),
        ('6_6_4_4_1', dict(targ=6, curr=6, updated=4, bat_size=4, min_srv=1,
                           batch=(6, 2))),
        ('6_4_0_6_1', dict(targ=6, curr=4, updated=0, bat_size=6, min_srv=1,
                           batch=(7, 6))),
        ('6_7_5_6_1', dict(targ=6, curr=7, updated=5, bat_size=6, min_srv=1,
                           batch=(6, 1))),
        ('6_4_0_4_2', dict(targ=6, curr=4, updated=0, bat_size=4, min_srv=2,
                           batch=(6, 4))),
        ('6_4_4_4_2', dict(targ=6, curr=4, updated=4, bat_size=4, min_srv=2,
                           batch=(6, 2))),
        ('6_4_0_4_4', dict(targ=6, curr=4, updated=0, bat_size=4, min_srv=4,
                           batch=(8, 4))),
        ('6_8_4_4_4', dict(targ=6, curr=4, updated=4, bat_size=4, min_srv=4,
                           batch=(6, 2))),
        ('6_8_2_4_4', dict(targ=6, curr=8, updated=2, bat_size=4, min_srv=4,
                           batch=(8, 4))),
        ('6_8_6_4_4', dict(targ=6, curr=8, updated=6, bat_size=4, min_srv=4,
                           batch=(6, 0))),
        ('6_4_0_5_6', dict(targ=6, curr=4, updated=0, bat_size=5, min_srv=6,
                           batch=(9, 5))),
        ('6_9_5_5_6', dict(targ=6, curr=9, updated=5, bat_size=5, min_srv=6,
                           batch=(7, 1))),
        ('6_9_2_5_6', dict(targ=6, curr=9, updated=2, bat_size=5, min_srv=6,
                           batch=(10, 4))),
        ('6_A_5_5_6', dict(targ=6, curr=10, updated=5, bat_size=5, min_srv=6,
                           batch=(7, 1))),
        ('6_7_6_5_6', dict(targ=6, curr=7, updated=6, bat_size=5, min_srv=6,
                           batch=(6, 0))),
    ]

    def test_next_batch(self):
        batch = rolling_update.next_batch(self.targ, self.curr,
                                          self.updated,
                                          self.bat_size,
                                          self.min_srv)
        self.assertEqual(self.batch, batch)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.common import netutils
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support


class SubnetPool(neutron.NeutronResource):
    """A resource that implements neutron subnet pool.

    This resource can be used to create a subnet pool with a large block
    of addresses and create subnets from it.
    """

    support_status = support.SupportStatus(version='6.0.0')

    required_service_extension = 'subnet_allocation'

    entity = 'subnetpool'

    PROPERTIES = (
        NAME, PREFIXES, ADDRESS_SCOPE, DEFAULT_QUOTA,
        DEFAULT_PREFIXLEN, MIN_PREFIXLEN, MAX_PREFIXLEN,
        IS_DEFAULT, TENANT_ID, SHARED, TAGS,
    ) = (
        'name', 'prefixes', 'address_scope', 'default_quota',
        'default_prefixlen', 'min_prefixlen', 'max_prefixlen',
        'is_default', 'tenant_id', 'shared', 'tags',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the subnet pool.'),
            update_allowed=True
        ),
        PREFIXES: properties.Schema(
            properties.Schema.LIST,
            _('List of subnet prefixes to assign.'),
            schema=properties.Schema(
                properties.Schema.STRING,
                constraints=[
                    constraints.CustomConstraint('net_cidr'),
                ],
            ),
            constraints=[constraints.Length(min=1)],
            required=True,
            update_allowed=True,
        ),
        ADDRESS_SCOPE: properties.Schema(
            properties.Schema.STRING,
            _('An address scope ID to assign to the subnet pool.'),
            constraints=[
                constraints.CustomConstraint('neutron.address_scope')
            ],
            update_allowed=True,
        ),
        DEFAULT_QUOTA: properties.Schema(
            properties.Schema.INTEGER,
            _('A per-tenant quota on the prefix space that can be allocated '
              'from the subnet pool for tenant subnets.'),
            constraints=[constraints.Range(min=0)],
            update_allowed=True,
        ),
        DEFAULT_PREFIXLEN: properties.Schema(
            properties.Schema.INTEGER,
            _('The size of the prefix to allocate when the cidr or '
              'prefixlen attributes are not specified while creating '
              'a subnet.'),
            constraints=[constraints.Range(min=0)],
            update_allowed=True,
        ),
        MIN_PREFIXLEN: properties.Schema(
            properties.Schema.INTEGER,
            _('Smallest prefix size that can be allocated '
              'from the subnet pool.'),
            constraints=[constraints.Range(min=0)],
            update_allowed=True,
            ),
        MAX_PREFIXLEN: properties.Schema(
            properties.Schema.INTEGER,
            _('Maximum prefix size that can be allocated '
              'from the subnet pool.'),
            constraints=[constraints.Range(min=0)],
            update_allowed=True,
        ),
        IS_DEFAULT: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether this is default IPv4/IPv6 subnet pool. '
              'There can only be one default subnet pool for each IP family. '
              'Note that the default policy setting restricts administrative '
              'users to set this to True.'),
            default=False,
            update_allowed=True,
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of the tenant who owns the subnet pool. Only '
              'administrative users can specify a tenant ID '
              'other than their own.')
        ),
        SHARED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether the subnet pool will be shared across all tenants. '
              'Note that the default policy setting restricts usage of this '
              'attribute to administrative users only.'),
            default=False,
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('The tags to be added to the subnetpool.'),
            schema=properties.Schema(properties.Schema.STRING),
            update_allowed=True,
            support_status=support.SupportStatus(version='9.0.0')
        ),
    }

    def validate(self):
        super(SubnetPool, self).validate()
        self._validate_prefix_bounds()

    def _validate_prefix_bounds(self):
        min_prefixlen = self.properties[self.MIN_PREFIXLEN]
        default_prefixlen = self.properties[self.DEFAULT_PREFIXLEN]
        max_prefixlen = self.properties[self.MAX_PREFIXLEN]
        msg_fmt = _('Illegal prefix bounds: %(key1)s=%(value1)s, '
                    '%(key2)s=%(value2)s.')
        # min_prefixlen can not be greater than max_prefixlen
        if min_prefixlen and max_prefixlen and min_prefixlen > max_prefixlen:
            msg = msg_fmt % dict(key1=self.MAX_PREFIXLEN,
                                 value1=max_prefixlen,
                                 key2=self.MIN_PREFIXLEN,
                                 value2=min_prefixlen)
            raise exception.StackValidationFailed(message=msg)

        if default_prefixlen:
            # default_prefixlen can not be greater than max_prefixlen
            if max_prefixlen and default_prefixlen > max_prefixlen:
                msg = msg_fmt % dict(key1=self.MAX_PREFIXLEN,
                                     value1=max_prefixlen,
                                     key2=self.DEFAULT_PREFIXLEN,
                                     value2=default_prefixlen)
                raise exception.StackValidationFailed(message=msg)
            # min_prefixlen can not be greater than default_prefixlen
            if min_prefixlen and min_prefixlen > default_prefixlen:
                msg = msg_fmt % dict(key1=self.MIN_PREFIXLEN,
                                     value1=min_prefixlen,
                                     key2=self.DEFAULT_PREFIXLEN,
                                     value2=default_prefixlen)
                raise exception.StackValidationFailed(message=msg)

    def _validate_prefixes_for_update(self, prop_diff):
        old_prefixes = self.properties[self.PREFIXES]
        new_prefixes = prop_diff[self.PREFIXES]
        # check new_prefixes is a superset of old_prefixes
        if not netutils.is_prefix_subset(old_prefixes, new_prefixes):
            msg = (_('Property %(key)s updated value %(new)s should '
                     'be superset of existing value '
                     '%(old)s.') % dict(key=self.PREFIXES,
                                        new=sorted(new_prefixes),
                                        old=sorted(old_prefixes)))
            raise exception.StackValidationFailed(message=msg)

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        if self.ADDRESS_SCOPE in props and props[self.ADDRESS_SCOPE]:
            client_plugin = self.client_plugin()
            scope_id = client_plugin.find_resourceid_by_name_or_id(
                client_plugin.RES_TYPE_ADDRESS_SCOPE,
                props.pop(self.ADDRESS_SCOPE))
            props['address_scope_id'] = scope_id
        tags = props.pop(self.TAGS, [])
        subnetpool = self.client().create_subnetpool(
            {'subnetpool': props})['subnetpool']
        self.resource_id_set(subnetpool['id'])

        if tags:
            self.set_tags(tags)

    def handle_delete(self):
        if self.resource_id is not None:
            with self.client_plugin().ignore_not_found:
                self.client().delete_subnetpool(self.resource_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        # check that new prefixes are superset of existing prefixes
        if self.PREFIXES in prop_diff:
            self._validate_prefixes_for_update(prop_diff)
        if self.ADDRESS_SCOPE in prop_diff:
            if prop_diff[self.ADDRESS_SCOPE]:
                client_plugin = self.client_plugin()
                scope_id = client_plugin.find_resourceid_by_name_or_id(
                    self.client(),
                    client_plugin.RES_TYPE_ADDRESS_SCOPE,
                    prop_diff.pop(self.ADDRESS_SCOPE))
            else:
                scope_id = prop_diff.pop(self.ADDRESS_SCOPE)
            prop_diff['address_scope_id'] = scope_id
        if self.TAGS in prop_diff:
            tags = prop_diff.pop(self.TAGS)
            self.set_tags(tags)
        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client().update_subnetpool(
                self.resource_id, {'subnetpool': prop_diff})


def resource_mapping():
    return {
        'OS::Neutron::SubnetPool': SubnetPool,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from webob import exc

from heat.api.openstack.v1 import util
from heat.common import context
from heat.common import policy
from heat.common import wsgi
from heat.tests import common


class TestGetAllowedParams(common.HeatTestCase):
    def setUp(self):
        super(TestGetAllowedParams, self).setUp()
        req = wsgi.Request({})
        self.params = req.params.copy()
        self.params.add('foo', 'foo value')
        self.param_types = {'foo': util.PARAM_TYPE_SINGLE}

    def test_returns_empty_dict(self):
        self.param_types = {}

        result = util.get_allowed_params(self.params, self.param_types)
        self.assertEqual({}, result)

    def test_only_adds_allowed_param_if_param_exists(self):
        self.param_types = {'foo': util.PARAM_TYPE_SINGLE}
        self.params.clear()

        result = util.get_allowed_params(self.params, self.param_types)
        self.assertNotIn('foo', result)

    def test_returns_only_allowed_params(self):
        self.params.add('bar', 'bar value')

        result = util.get_allowed_params(self.params, self.param_types)
        self.assertIn('foo', result)
        self.assertNotIn('bar', result)

    def test_handles_single_value_params(self):
        result = util.get_allowed_params(self.params, self.param_types)
        self.assertEqual('foo value', result['foo'])

    def test_handles_multiple_value_params(self):
        self.param_types = {'foo': util.PARAM_TYPE_MULTI}
        self.params.add('foo', 'foo value 2')

        result = util.get_allowed_params(self.params, self.param_types)
        self.assertEqual(2, len(result['foo']))
        self.assertIn('foo value', result['foo'])
        self.assertIn('foo value 2', result['foo'])

    def test_handles_mixed_value_param_with_multiple_entries(self):
        self.param_types = {'foo': util.PARAM_TYPE_MIXED}
        self.params.add('foo', 'foo value 2')

        result = util.get_allowed_params(self.params, self.param_types)
        self.assertEqual(2, len(result['foo']))
        self.assertIn('foo value', result['foo'])
        self.assertIn('foo value 2', result['foo'])

    def test_handles_mixed_value_param_with_single_entry(self):
        self.param_types = {'foo': util.PARAM_TYPE_MIXED}

        result = util.get_allowed_params(self.params, self.param_types)
        self.assertEqual('foo value', result['foo'])

    def test_bogus_param_type(self):
        self.param_types = {'foo': 'blah'}
        self.assertRaises(AssertionError, util.get_allowed_params,
                          self.params, self.param_types)


class TestPolicyEnforce(common.HeatTestCase):
    def setUp(self):
        super(TestPolicyEnforce, self).setUp()
        self.req = wsgi.Request({})
        self.req.context = context.RequestContext(project_id='foo',
                                                  is_admin=False)

        class DummyController(object):
            REQUEST_SCOPE = 'test'

            @util.registered_policy_enforce
            def an_action(self, req):
                return 'woot'

        self.controller = DummyController()

    @mock.patch.object(policy.Enforcer, 'enforce')
    def test_policy_enforce_tenant_mismatch(self, mock_enforce):
        mock_enforce.return_value = True

        self.assertEqual('woot',
                         self.controller.an_action(self.req, 'foo'))

        self.assertRaises(exc.HTTPForbidden,
                          self.controller.an_action,
                          self.req, tenant_id='bar')

    @mock.patch.object(policy.Enforcer, 'enforce')
    def test_policy_enforce_tenant_mismatch_is_admin(self, mock_enforce):
        self.req.context = context.RequestContext(project_id='foo',
                                                  is_admin=True)
        mock_enforce.return_value = True

        self.assertEqual('woot',
                         self.controller.an_action(self.req, 'foo'))

        self.assertEqual('woot',
                         self.controller.an_action(self.req, 'bar'))

    @mock.patch.object(policy.Enforcer, 'enforce')
    def test_policy_enforce_policy_deny(self, mock_enforce):
        mock_enforce.return_value = False

        self.assertRaises(exc.HTTPForbidden,
                          self.controller.an_action,
                          self.req, tenant_id='foo')



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from heat.rpc import worker_api as rpc_api
from heat.rpc import worker_client as rpc_client
from heat.tests import common


class WorkerClientTest(common.HeatTestCase):

    def setUp(self):
        super(WorkerClientTest, self).setUp()
        self.fake_engine_id = 'fake-engine-id'

    def test_make_msg(self):
        method = 'sample_method'
        kwargs = {'a': '1',
                  'b': '2'}
        result = method, kwargs
        self.assertEqual(
            result,
            rpc_client.WorkerClient.make_msg(method, **kwargs))

    @mock.patch('heat.common.messaging.get_rpc_client',
                return_value=mock.Mock())
    def test_cast(self, rpc_client_method):
        # Mock the rpc client
        mock_rpc_client = rpc_client_method.return_value

        # Create the WorkerClient
        worker_client = rpc_client.WorkerClient()
        rpc_client_method.assert_called_once_with(
            version=rpc_client.WorkerClient.BASE_RPC_API_VERSION,
            topic=rpc_api.TOPIC)

        self.assertEqual(mock_rpc_client,
                         worker_client._client,
                         "Failed to create RPC client")

        # Check cast in default version
        mock_cnxt = mock.Mock()
        method = 'sample_method'
        kwargs = {'a': '1',
                  'b': '2'}
        msg = method, kwargs

        # go with default version
        return_value = worker_client.cast(mock_cnxt, msg)
        self.assertIsNone(return_value)
        mock_rpc_client.cast.assert_called_with(mock_cnxt,
                                                method,
                                                **kwargs)

        # Check cast in given version
        version = '1.3'
        return_value = worker_client.cast(mock_cnxt, msg, version)
        self.assertIsNone(return_value)
        mock_rpc_client.prepare.assert_called_once_with(version=version)
        mock_rpc_client.cast.assert_called_once_with(mock_cnxt,
                                                     method,
                                                     **kwargs)

    def test_cancel_check_resource(self):
        mock_stack_id = 'dummy-stack-id'
        mock_cnxt = mock.Mock()
        method = 'cancel_check_resource'
        kwargs = {'stack_id': mock_stack_id}
        mock_rpc_client = mock.MagicMock()
        mock_cast = mock.MagicMock()
        with mock.patch('heat.common.messaging.get_rpc_client') as mock_grc:
            mock_grc.return_value = mock_rpc_client
            mock_rpc_client.prepare.return_value = mock_cast
            wc = rpc_client.WorkerClient()
            ret_val = wc.cancel_check_resource(mock_cnxt, mock_stack_id,
                                               self.fake_engine_id)
            # ensure called with fanout=True
            mock_grc.assert_called_with(
                version=wc.BASE_RPC_API_VERSION,
                topic=rpc_api.TOPIC,
                server=self.fake_engine_id)
            self.assertIsNone(ret_val)
            mock_rpc_client.prepare.assert_called_with(
                version='1.3')
            # ensure correct rpc method is called
            mock_cast.cast.assert_called_with(mock_cnxt, method, **kwargs)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.engine import dependencies
from heat.tests import common


class dependenciesTest(common.HeatTestCase):

    def _dep_test(self, func, checkorder, deps):
        nodes = set.union(*[set(e) for e in deps])

        d = dependencies.Dependencies(deps)
        order = list(func(d))

        for n in nodes:
            self.assertIn(n, order, '"%s" is not in the sequence' % n)
            self.assertEqual(1, order.count(n))

        self.assertEqual(len(nodes), len(order))

        for lr, fr in deps:
            checkorder(order.index(fr), order.index(lr))

    def _dep_test_fwd(self, *deps):
        def assertLess(a, b):
            self.assertTrue(a < b,
                            '"%s" is not less than "%s"' % (str(a), str(b)))
        self._dep_test(iter, assertLess, deps)

    def _dep_test_rev(self, *deps):
        def assertGreater(a, b):
            self.assertTrue(a > b,
                            '"%s" is not greater than "%s"' % (str(a), str(b)))
        self._dep_test(reversed, assertGreater, deps)

    def test_edges(self):
        input_edges = [('1', None), ('2', '3'), ('2', '4')]
        dp = dependencies.Dependencies(input_edges)
        self.assertEqual(set(input_edges), set(dp.graph().edges()))

    def test_repr(self):
        dp = dependencies.Dependencies([('1', None), ('2', '3'), ('2', '4')])
        s = "Dependencies([('1', None), ('2', '3'), ('2', '4')])"
        self.assertEqual(s, repr(dp))

    def test_single_node(self):
        d = dependencies.Dependencies([('only', None)])
        li = list(iter(d))
        self.assertEqual(1, len(li))
        self.assertEqual('only', li[0])

    def test_disjoint(self):
        d = dependencies.Dependencies([('1', None), ('2', None)])
        li = list(iter(d))
        self.assertEqual(2, len(li))
        self.assertIn('1', li)
        self.assertIn('2', li)

    def test_single_fwd(self):
        self._dep_test_fwd(('second', 'first'))

    def test_single_rev(self):
        self._dep_test_rev(('second', 'first'))

    def test_chain_fwd(self):
        self._dep_test_fwd(('third', 'second'), ('second', 'first'))

    def test_chain_rev(self):
        self._dep_test_rev(('third', 'second'), ('second', 'first'))

    def test_diamond_fwd(self):
        self._dep_test_fwd(('last', 'mid1'), ('last', 'mid2'),
                           ('mid1', 'first'), ('mid2', 'first'))

    def test_diamond_rev(self):
        self._dep_test_rev(('last', 'mid1'), ('last', 'mid2'),
                           ('mid1', 'first'), ('mid2', 'first'))

    def test_complex_fwd(self):
        self._dep_test_fwd(('last', 'mid1'), ('last', 'mid2'),
                           ('mid1', 'mid3'), ('mid1', 'first'),
                           ('mid3', 'first'), ('mid2', 'first'))

    def test_complex_rev(self):
        self._dep_test_rev(('last', 'mid1'), ('last', 'mid2'),
                           ('mid1', 'mid3'), ('mid1', 'first'),
                           ('mid3', 'first'), ('mid2', 'first'))

    def test_many_edges_fwd(self):
        self._dep_test_fwd(('last', 'e1'), ('last', 'mid1'), ('last', 'mid2'),
                           ('mid1', 'e2'), ('mid1', 'mid3'),
                           ('mid2', 'mid3'),
                           ('mid3', 'e3'))

    def test_many_edges_rev(self):
        self._dep_test_rev(('last', 'e1'), ('last', 'mid1'), ('last', 'mid2'),
                           ('mid1', 'e2'), ('mid1', 'mid3'),
                           ('mid2', 'mid3'),
                           ('mid3', 'e3'))

    def test_dbldiamond_fwd(self):
        self._dep_test_fwd(('last', 'a1'), ('last', 'a2'),
                           ('a1', 'b1'), ('a2', 'b1'), ('a2', 'b2'),
                           ('b1', 'first'), ('b2', 'first'))

    def test_dbldiamond_rev(self):
        self._dep_test_rev(('last', 'a1'), ('last', 'a2'),
                           ('a1', 'b1'), ('a2', 'b1'), ('a2', 'b2'),
                           ('b1', 'first'), ('b2', 'first'))

    def test_circular_fwd(self):
        d = dependencies.Dependencies([('first', 'second'),
                                       ('second', 'third'),
                                       ('third', 'first')])
        self.assertRaises(exception.CircularDependencyException,
                          list,
                          iter(d))

    def test_circular_rev(self):
        d = dependencies.Dependencies([('first', 'second'),
                                       ('second', 'third'),
                                       ('third', 'first')])
        self.assertRaises(exception.CircularDependencyException,
                          list,
                          reversed(d))

    def test_self_ref(self):
        d = dependencies.Dependencies([('node', 'node')])
        self.assertRaises(exception.CircularDependencyException,
                          list,
                          iter(d))

    def test_complex_circular_fwd(self):
        d = dependencies.Dependencies([('last', 'e1'), ('last', 'mid1'),
                                       ('last', 'mid2'), ('mid1', 'e2'),
                                       ('mid1', 'mid3'), ('mid2', 'mid3'),
                                       ('mid3', 'e3'), ('e3', 'mid1')])
        self.assertRaises(exception.CircularDependencyException,
                          list,
                          iter(d))

    def test_complex_circular_rev(self):
        d = dependencies.Dependencies([('last', 'e1'), ('last', 'mid1'),
                                       ('last', 'mid2'), ('mid1', 'e2'),
                                       ('mid1', 'mid3'), ('mid2', 'mid3'),
                                       ('mid3', 'e3'), ('e3', 'mid1')])
        self.assertRaises(exception.CircularDependencyException,
                          list,
                          reversed(d))

    def test_noexist_partial(self):
        d = dependencies.Dependencies([('foo', 'bar')])

        def get(i):
            return d[i]
        self.assertRaises(KeyError, get, 'baz')

    def test_single_partial(self):
        d = dependencies.Dependencies([('last', 'first')])
        p = d['last']
        li = list(iter(p))
        self.assertEqual(1, len(li))
        self.assertEqual('last', li[0])

    def test_simple_partial(self):
        d = dependencies.Dependencies([('last', 'middle'),
                                       ('middle', 'first')])
        p = d['middle']
        order = list(iter(p))
        self.assertEqual(2, len(order))
        for n in ('last', 'middle'):
            self.assertIn(n, order,
                          "'%s' not found in dependency order" % n)
        self.assertGreater(order.index('last'), order.index('middle'))

    def test_simple_multilevel_partial(self):
        d = dependencies.Dependencies([('last', 'middle'),
                                       ('middle', 'target'),
                                       ('target', 'first')])
        p = d['target']
        order = list(iter(p))
        self.assertEqual(3, len(order))
        for n in ('last', 'middle', 'target'):
            self.assertIn(n, order,
                          "'%s' not found in dependency order" % n)

    def test_complex_partial(self):
        d = dependencies.Dependencies([('last', 'e1'), ('last', 'mid1'),
                                       ('last', 'mid2'), ('mid1', 'e2'),
                                       ('mid1', 'mid3'), ('mid2', 'mid3'),
                                       ('mid3', 'e3')])
        p = d['mid3']
        order = list(iter(p))
        self.assertEqual(4, len(order))
        for n in ('last', 'mid1', 'mid2', 'mid3'):
            self.assertIn(n, order,
                          "'%s' not found in dependency order" % n)

    def test_required_by(self):
        d = dependencies.Dependencies([('last', 'e1'), ('last', 'mid1'),
                                       ('last', 'mid2'), ('mid1', 'e2'),
                                       ('mid1', 'mid3'), ('mid2', 'mid3'),
                                       ('mid3', 'e3')])

        self.assertEqual(0, len(list(d.required_by('last'))))

        required_by = list(d.required_by('mid3'))
        self.assertEqual(2, len(required_by))
        for n in ('mid1', 'mid2'):
            self.assertIn(n, required_by,
                          "'%s' not found in required_by" % n)

        required_by = list(d.required_by('e2'))
        self.assertEqual(1, len(required_by))
        self.assertIn('mid1', required_by,
                      "'%s' not found in required_by" % n)

        self.assertRaises(KeyError, d.required_by, 'foo')

    def test_leaves(self):
        d = dependencies.Dependencies([('last1', 'mid'), ('last2', 'mid'),
                                       ('mid', 'first1'), ('mid', 'first2')])

        leaves = sorted(list(d.leaves()))

        self.assertEqual(['first1', 'first2'], leaves)

    def test_roots(self):
        d = dependencies.Dependencies([('last1', 'mid'), ('last2', 'mid'),
                                       ('mid', 'first1'), ('mid', 'first2')])

        leaves = sorted(list(d.roots()))

        self.assertEqual(['last1', 'last2'], leaves)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine import template_files
from heat.tests import common
from heat.tests import utils

template_files_1 = {'template file 1': 'Contents of template 1',
                    'template file 2': 'More template contents'}


class TestTemplateFiles(common.HeatTestCase):

    def test_cache_miss(self):
        ctx = utils.dummy_context()
        tf1 = template_files.TemplateFiles(template_files_1)
        tf1.store(ctx)
        # As this is the only reference to the value in _d, deleting
        # t1.files will cause the value to get removed from _d (due to
        # it being a WeakValueDictionary.
        del tf1.files
        self.assertNotIn(tf1.files_id, template_files._d)
        # this will cause the cache refresh
        self.assertEqual(template_files_1['template file 1'],
                         tf1['template file 1'])
        self.assertEqual(template_files_1, template_files._d[tf1.files_id])

    def test_d_weakref_behaviour(self):
        ctx = utils.dummy_context()
        tf1 = template_files.TemplateFiles(template_files_1)
        tf1.store(ctx)
        tf2 = template_files.TemplateFiles(tf1)
        del tf1.files
        self.assertIn(tf2.files_id, template_files._d)
        del tf2.files
        self.assertNotIn(tf2.files_id, template_files._d)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class Segment(neutron.NeutronResource):
    """A resource for Neutron Segment.

    This requires enabling the segments service plug-in by appending
    'segments' to the list of service_plugins in the neutron.conf.

    The default policy usage of this resource is limited to
    administrators only.
    """

    required_service_extension = 'segment'

    support_status = support.SupportStatus(version='9.0.0')

    NETWORK_TYPES = [LOCAL, VLAN, VXLAN,
                     GRE, GENEVE, FLAT] = ['local', 'vlan', 'vxlan',
                                           'gre', 'geneve', 'flat']
    PROPERTIES = (
        NETWORK, NETWORK_TYPE, PHYSICAL_NETWORK, SEGMENTATION_ID,
        NAME, DESCRIPTION
    ) = (
        'network', 'network_type', 'physical_network', 'segmentation_id',
        'name', 'description'
    )

    properties_schema = {
        NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('The name/id of network to associate with this segment.'),
            constraints=[constraints.CustomConstraint('neutron.network')],
            required=True
        ),
        NETWORK_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Type of network to associate with this segment.'),
            constraints=[
                constraints.AllowedValues(NETWORK_TYPES),
            ],
            required=True
        ),
        PHYSICAL_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('Name of physical network to associate with this segment.'),
        ),
        SEGMENTATION_ID: properties.Schema(
            properties.Schema.INTEGER,
            _('Segmentation ID for this segment.'),
            constraints=[
                constraints.Range(min=1),
            ],
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the segment.'),
            update_allowed=True,
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of the segment.'),
            update_allowed=True,
        ),
    }

    def translation_rules(self, props):
        client_plugin = self.client_plugin()
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.NETWORK],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_NETWORK
            )]

    def validate(self):
        super(Segment, self).validate()
        phys_network = self.properties[self.PHYSICAL_NETWORK]
        network_type = self.properties[self.NETWORK_TYPE]
        seg_id = self.properties[self.SEGMENTATION_ID]

        msg_fmt = _('%(prop)s is required for %(type)s provider network.')
        if network_type in [self.FLAT, self.VLAN] and phys_network is None:
            msg = msg_fmt % {'prop': self.PHYSICAL_NETWORK,
                             'type': network_type}
            raise exception.StackValidationFailed(message=msg)

        if network_type == self.VLAN and seg_id is None:
            msg = msg_fmt % {'prop': self.SEGMENTATION_ID,
                             'type': network_type}
            raise exception.StackValidationFailed(message=msg)

        msg_fmt = _('%(prop)s is prohibited for %(type)s provider network.')
        if network_type in [self.LOCAL, self.FLAT] and seg_id is not None:
            msg = msg_fmt % {'prop': self.SEGMENTATION_ID,
                             'type': network_type}
            raise exception.StackValidationFailed(message=msg)

        tunnel_types = [self.VXLAN, self.GRE, self.GENEVE]
        if network_type in tunnel_types and phys_network is not None:
            msg = msg_fmt % {'prop': self.PHYSICAL_NETWORK,
                             'type': network_type}
            raise exception.StackValidationFailed(message=msg)

        if network_type == self.VLAN and seg_id > 4094:
            msg = _('Up to 4094 VLAN network segments can exist '
                    'on each physical_network.')
            raise exception.StackValidationFailed(message=msg)

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        props['network_id'] = props.pop(self.NETWORK)
        segment = self.client('openstack').network.create_segment(**props)
        self.resource_id_set(segment.id)

    def handle_delete(self):
        if self.resource_id is None:
            return

        with self.client_plugin('openstack').ignore_not_found:
            self.client('openstack').network.delete_segment(self.resource_id)

    def needs_replace_failed(self):
        if not self.resource_id:
            return True

        with self.client_plugin('openstack').ignore_not_found:
            self._show_resource()
            return False

        return True

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client('openstack').network.update_segment(
                self.resource_id, **prop_diff)

    def _show_resource(self):
        return self.client('openstack').network.get_segment(
            self.resource_id).to_dict()


def resource_mapping():
    return {
        'OS::Neutron::Segment': Segment
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class FlavorProfile(resource.Resource):
    """A resource for creating octavia Flavor Profiles.

    This resource creates and manages octavia Flavor Profiles,
    which allows to tune Load Balancers' capabilities.
    """

    default_client_name = 'octavia'

    support_status = support.SupportStatus(version='14.0.0')

    PROPERTIES = (
        NAME, FLAVOR_DATA, PROVIDER_NAME
    ) = (
        'name', 'flavor_data', 'provider_name'
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of this Flavor Profile.'),
            update_allowed=True
        ),
        FLAVOR_DATA: properties.Schema(
            properties.Schema.STRING,
            _('JSON string containing the flavor metadata.'),
            update_allowed=True,
            required=True,
            constraints=[constraints.CustomConstraint('json_string')]
        ),
        PROVIDER_NAME: properties.Schema(
            properties.Schema.STRING,
            _('Provider name of this Flavor Profile.'),
            update_allowed=True,
        ),
    }

    def _prepare_args(self, properties):
        props = dict((k, v) for k, v in properties.items()
                     if v is not None)
        if self.NAME not in props:
            props[self.NAME] = self.physical_resource_name()
        return props

    def handle_create(self):
        props = self._prepare_args(self.properties)

        flavorprofile = self.client().flavorprofile_create(
            json={'flavorprofile': props})['flavorprofile']
        self.resource_id_set(flavorprofile['id'])

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            if self.NAME in prop_diff and prop_diff[self.NAME] is None:
                prop_diff[self.NAME] = self.physical_resource_name()
            self.client().flavorprofile_set(
                self.resource_id,
                json={'flavorprofile': prop_diff})

    def handle_delete(self):
        with self.client_plugin().ignore_not_found:
            self.client().flavorprofile_delete(self.resource_id)
            return True

    def _show_resource(self):
        return self.client().flavorprofile_show(self.resource_id)


def resource_mapping():
    return {
        'OS::Octavia::FlavorProfile': FlavorProfile
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from webob import exc

from heat.common.i18n import _
from heat.common import identifier


def registered_policy_enforce(handler):
    """Decorator that enforces policies.

    Checks the path matches the request context and enforce policy defined in
    policies.

    This is a handler method decorator.
    """
    @functools.wraps(handler)
    def handle_stack_method(controller, req, tenant_id, **kwargs):
        # NOTE(tkajinam): Heat uses stack owner's project id in redirect URI.
        # So admin might be redirected to different project id when accessing
        # resources in a different project. Use project id in context to
        # bypass project_id check, because admin should have access to all
        # projects.
        if req.context.is_admin and req.context.project_id:
            tenant_id = req.context.project_id
        _target = {"project_id": tenant_id}
        if req.context.project_id != tenant_id:
            raise exc.HTTPForbidden()
        allowed = req.context.policy.enforce(
            context=req.context,
            action=handler.__name__,
            scope=controller.REQUEST_SCOPE,
            target=_target,
            is_registered_policy=True)
        if not allowed:
            raise exc.HTTPForbidden()
        return handler(controller, req, **kwargs)

    return handle_stack_method


def no_policy_enforce(handler):
    """Decorator that does *not* enforce policies.

    Checks the path matches the request context.

    This is a handler method decorator.
    """
    @functools.wraps(handler)
    def handle_stack_method(controller, req, tenant_id, **kwargs):
        if req.context.project_id != tenant_id and not req.context.is_admin:
            raise exc.HTTPForbidden()
        return handler(controller, req, **kwargs)

    return handle_stack_method


def registered_identified_stack(handler):
    """Decorator that passes a stack identifier instead of path components.

    This is a handler method decorator. Policy is enforced using a registered
    policy name.
    """
    return registered_policy_enforce(_identified_stack(handler))


def _identified_stack(handler):
    @functools.wraps(handler)
    def handle_stack_method(controller, req, stack_name, stack_id, **kwargs):
        stack_identity = identifier.HeatIdentifier(req.context.project_id,
                                                   stack_name,
                                                   stack_id)
        return handler(controller, req, dict(stack_identity), **kwargs)

    return handle_stack_method


def make_url(req, identity):
    """Return the URL for the supplied identity dictionary."""
    try:
        stack_identity = identifier.HeatIdentifier(**identity)
    except ValueError:
        err_reason = _('Invalid Stack address')
        raise exc.HTTPInternalServerError(err_reason)

    return req.relative_url(stack_identity.url_path(), True)


def make_link(req, identity, relationship='self'):
    """Return a link structure for the supplied identity dictionary."""
    return {'href': make_url(req, identity), 'rel': relationship}


PARAM_TYPES = (
    PARAM_TYPE_SINGLE, PARAM_TYPE_MULTI, PARAM_TYPE_MIXED
) = (
    'single', 'multi', 'mixed'
)


def get_allowed_params(params, param_types):
    """Extract from ``params`` all entries listed in ``param_types``.

    The returning dict will contain an entry for a key if, and only if,
    there's an entry in ``param_types`` for that key and at least one entry in
    ``params``. If ``params`` contains multiple entries for the same key, it
    will yield an array of values: ``{key: [v1, v2,...]}``

    :param params: a NestedMultiDict from webob.Request.params
    :param param_types: an dict of allowed parameters and their types

    :returns: a dict with {key: value} pairs
    """
    allowed_params = {}

    for key, get_type in param_types.items():
        assert get_type in PARAM_TYPES

        value = None
        if get_type == PARAM_TYPE_SINGLE:
            value = params.get(key)
        elif get_type == PARAM_TYPE_MULTI:
            value = params.getall(key)
        elif get_type == PARAM_TYPE_MIXED:
            value = params.getall(key)
            if isinstance(value, list) and len(value) == 1:
                value = value.pop()

        if value:
            allowed_params[key] = value

    return allowed_params



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class KeystoneProject(resource.Resource):
    """Heat Template Resource for Keystone Project.

    Projects represent the base unit of ownership in OpenStack, in that all
    resources in OpenStack should be owned by a specific project. A project
    itself must be owned by a specific domain, and hence all project names are
    not globally unique, but unique to their domain. If the domain for a
    project is not specified, then it is added to the default domain.
    """

    support_status = support.SupportStatus(
        version='2015.1',
        message=_('Supported versions: keystone v3'))

    default_client_name = 'keystone'

    entity = 'projects'

    PROPERTIES = (
        NAME, DOMAIN, DESCRIPTION, ENABLED, PARENT, TAGS,
    ) = (
        'name', 'domain', 'description', 'enabled', 'parent', 'tags',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of keystone project.'),
            update_allowed=True
        ),
        DOMAIN: properties.Schema(
            properties.Schema.STRING,
            _('Name or id of keystone domain.'),
            default='default',
            update_allowed=True,
            constraints=[constraints.CustomConstraint('keystone.domain')]
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of keystone project.'),
            default='',
            update_allowed=True
        ),
        ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('This project is enabled or disabled.'),
            default=True,
            update_allowed=True
        ),
        PARENT: properties.Schema(
            properties.Schema.STRING,
            _('The name or ID of parent of this keystone project '
              'in hierarchy.'),
            support_status=support.SupportStatus(version='6.0.0'),
            constraints=[constraints.CustomConstraint('keystone.project')]
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('A list of tags for labeling and sorting projects.'),
            support_status=support.SupportStatus(version='10.0.0'),
            default=[],
            update_allowed=True
        ),
    }

    ATTRIBUTES = (
        NAME_ATTR, PARENT_ATTR, DOMAIN_ATTR, ENABLED_ATTR, IS_DOMAIN_ATTR
    ) = (
        'name', 'parent_id', 'domain_id', 'enabled', 'is_domain'
    )
    attributes_schema = {
        NAME_ATTR: attributes.Schema(
            _('Project name.'),
            support_status=support.SupportStatus(version='10.0.0'),
            type=attributes.Schema.STRING
        ),
        PARENT_ATTR: attributes.Schema(
            _('Parent project id.'),
            support_status=support.SupportStatus(version='10.0.0'),
            type=attributes.Schema.STRING
        ),
        DOMAIN_ATTR: attributes.Schema(
            _('Domain id for project.'),
            support_status=support.SupportStatus(version='10.0.0'),
            type=attributes.Schema.STRING
        ),
        ENABLED_ATTR: attributes.Schema(
            _('Flag of enable project.'),
            support_status=support.SupportStatus(version='10.0.0'),
            type=attributes.Schema.BOOLEAN
        ),
        IS_DOMAIN_ATTR: attributes.Schema(
            _('Indicates whether the project also acts as a domain.'),
            support_status=support.SupportStatus(version='10.0.0'),
            type=attributes.Schema.BOOLEAN
        ),
    }

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        project = self.client().projects.get(self.resource_id)
        return getattr(project, name, None)

    def translation_rules(self, properties):
        return [
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                [self.DOMAIN],
                client_plugin=self.client_plugin(),
                finder='get_domain_id'
            ),
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                [self.PARENT],
                client_plugin=self.client_plugin(),
                finder='get_project_id'
            ),
        ]

    def client(self):
        return super(KeystoneProject, self).client().client

    def handle_create(self):
        project_name = (self.properties[self.NAME] or
                        self.physical_resource_name())
        description = self.properties[self.DESCRIPTION]
        domain = self.properties[self.DOMAIN]
        enabled = self.properties[self.ENABLED]
        parent = self.properties[self.PARENT]
        tags = self.properties[self.TAGS]

        project = self.client().projects.create(
            name=project_name,
            domain=domain,
            description=description,
            enabled=enabled,
            parent=parent,
            tags=tags)

        self.resource_id_set(project.id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            name = None
            # Don't update the name if no change
            if self.NAME in prop_diff:
                name = prop_diff[self.NAME] or self.physical_resource_name()

            description = prop_diff.get(self.DESCRIPTION)
            enabled = prop_diff.get(self.ENABLED)
            domain = prop_diff.get(self.DOMAIN, self.properties[self.DOMAIN])
            tags = (prop_diff.get(self.TAGS) or
                    self.properties[self.TAGS])

            self.client().projects.update(
                project=self.resource_id,
                name=name,
                description=description,
                enabled=enabled,
                domain=domain,
                tags=tags
            )

    def parse_live_resource_data(self, resource_properties, resource_data):
        result = super(KeystoneProject, self).parse_live_resource_data(
            resource_properties, resource_data)
        result[self.DOMAIN] = resource_data.get('domain_id')
        return result

    def handle_delete(self):
        if self.resource_id:
            # find and delete the default security group Neutron has created
            default_sec_group_name = "default"
            nclient = self.client_plugin("neutron").client()
            default_sec_groups = nclient.list_security_groups(
                project_id=self.resource_id,
                name=default_sec_group_name)["security_groups"]
            # NOTE(pas-ha) this should always contain a single security group
            # (if any) as Netron enforces uniqueness of 'default' security
            # group in a project.
            # However leaving orphans is bad enough, so we are deleting
            # any security group with such name w/o uniqueness check.
            for secgroup in default_sec_groups:
                with self.client_plugin("neutron").ignore_not_found:
                    nclient.delete_security_group(secgroup["id"])
        super(KeystoneProject, self).handle_delete()


def resource_mapping():
    return {
        'OS::Keystone::Project': KeystoneProject
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import keystoneauth1.exceptions as kc_exception
from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine import resource

LOG = logging.getLogger(__name__)


class StackUser(resource.Resource):

    # Subclasses create a user, and optionally keypair associated with a
    # resource in a stack. Users are created in the heat stack user domain
    # (in a project specific to the stack)
    def handle_create(self):
        self._create_user()

    def _create_user(self):
        if self.data().get('user_id'):
            # a user has been created already
            return

        # Check for stack user project, create if not yet set
        if not self.stack.stack_user_project_id:
            project_id = self.keystone().create_stack_domain_project(
                self.stack.id)
            self.stack.set_stack_user_project_id(project_id)

        # Create a keystone user in the stack domain project
        user_id = self.keystone().create_stack_domain_user(
            username=self.physical_resource_name(),
            password=getattr(self, 'password', None),
            project_id=self.stack.stack_user_project_id)

        # Store the ID in resource data, for compatibility with SignalResponder
        self.data_set('user_id', user_id)

    def _user_token(self):
        project_id = self.stack.stack_user_project_id
        if not project_id:
            raise ValueError(_("Can't get user token, user not yet created"))
        password = getattr(self, 'password', None)
        # FIXME(shardy): the create and getattr here could allow insane
        # passwords, e.g a zero length string, if these happen it almost
        # certainly means a bug elsewhere in heat, so add assertion to catch
        if password is None:
            raise ValueError(_("Can't get user token without password"))

        return self.keystone().stack_domain_user_token(
            user_id=self._get_user_id(),
            project_id=project_id, password=password)

    def _get_user_id(self):
        user_id = self.data().get('user_id')
        if user_id:
            return user_id

    def handle_delete(self):
        self._delete_user()

        return super(StackUser, self).handle_delete()

    def _delete_user(self):
        user_id = self._get_user_id()
        if user_id is None:
            return

        # the user is going away, so we want the keypair gone as well
        self._delete_keypair()

        try:
            self.keystone().delete_stack_domain_user(
                user_id=user_id, project_id=self.stack.stack_user_project_id)
        except kc_exception.NotFound:
            pass
        except ValueError:
            # FIXME(shardy): This is a legacy delete path for backwards
            # compatibility with resources created before the migration
            # to stack_user.StackUser domain users.  After an appropriate
            # transitional period, this should be removed.
            LOG.warning('Reverting to legacy user delete path')
            try:
                self.keystone().delete_stack_user(user_id)
            except kc_exception.NotFound:
                pass
        self.data_delete('user_id')

    def handle_suspend(self):
        user_id = self._get_user_id()
        try:
            self.keystone().disable_stack_domain_user(
                user_id=user_id, project_id=self.stack.stack_user_project_id)
        except ValueError:
            # FIXME(shardy): This is a legacy path for backwards compatibility
            self.keystone().disable_stack_user(user_id=user_id)

    def handle_resume(self):
        user_id = self._get_user_id()
        try:
            self.keystone().enable_stack_domain_user(
                user_id=user_id, project_id=self.stack.stack_user_project_id)
        except ValueError:
            # FIXME(shardy): This is a legacy path for backwards compatibility
            self.keystone().enable_stack_user(user_id=user_id)

    def _create_keypair(self):
        # Subclasses may optionally call this in handle_create to create
        # an ec2 keypair associated with the user, the resulting keys are
        # stored in resource_data
        if self.data().get('credential_id'):
            return  # a keypair was created already

        user_id = self._get_user_id()
        kp = self.keystone().create_stack_domain_user_keypair(
            user_id=user_id, project_id=self.stack.stack_user_project_id)
        if not kp:
            raise exception.Error(_("Error creating ec2 keypair for user %s") %
                                  user_id)
        else:
            try:
                credential_id = kp.id
            except AttributeError:
                # keystone v2 keypairs do not have an id attribute. Use the
                # access key instead.
                credential_id = kp.access
            self.data_set('credential_id', credential_id, redact=True)
            self.data_set('access_key', kp.access, redact=True)
            self.data_set('secret_key', kp.secret, redact=True)
        return kp

    def _delete_keypair(self):
        # Subclasses may optionally call this to delete a keypair created
        # via _create_keypair
        credential_id = self.data().get('credential_id')
        if not credential_id:
            return
        user_id = self._get_user_id()
        if user_id is None:
            return

        try:
            self.keystone().delete_stack_domain_user_keypair(
                user_id=user_id, project_id=self.stack.stack_user_project_id,
                credential_id=credential_id)
        except kc_exception.NotFound:
            pass
        except ValueError:
            self.keystone().delete_ec2_keypair(
                user_id=user_id, credential_id=credential_id)

        for data_key in ('access_key', 'secret_key', 'credential_id'):
            self.data_delete(data_key)

    def _register_access_key(self):
        """Access is limited to this resource, which created the keypair."""
        def access_allowed(resource_name):
            return resource_name == self.name

        if self.access_key is not None:
            self.stack.register_access_allowed_handler(
                self.access_key, access_allowed)
        if self._get_user_id() is not None:
            self.stack.register_access_allowed_handler(
                self._get_user_id(), access_allowed)
